@@ -43,6 +43,7 @@
 #include "../libsnap-confine-private/cleanup-funcs.h"
 #include "../libsnap-confine-private/mount-opt.h"
 #include "../libsnap-confine-private/mountinfo.h"
+#include "../libsnap-confine-private/snap-dir.h"
 #include "../libsnap-confine-private/snap.h"
 #include "../libsnap-confine-private/string-utils.h"
 #include "../libsnap-confine-private/tool.h"
@@ -111,7 +112,7 @@ static void setup_private_tmp(const char *snap_instance)
 		die("%s has unexpected ownership / permissions",
 		    SNAP_PRIVATE_TMP_ROOT_DIR);
 	}
-	// Create /tmp/snap-private-tmp/snap.$SNAP_INSTANCE_NAME/ 0700 root.root.
+	// Create /tmp/snap-private-tmp/snap.$SNAP_INSTANCE_NAME/ 0700 root:root.
 	sc_must_snprintf(base, sizeof(base), "snap.%s", snap_instance);
 	if (mkdirat(private_tmp_root_fd, base, 0700) < 0 && errno != EEXIST) {
 		die("cannot create base directory: %s", base);
@@ -129,7 +130,7 @@ static void setup_private_tmp(const char *snap_instance)
 		die("%s/%s has unexpected ownership / permissions",
 		    SNAP_PRIVATE_TMP_ROOT_DIR, base);
 	}
-	// Create /tmp/$PRIVATE/snap.$SNAP_NAME/tmp 01777 root.root Ignore EEXIST since we
+	// Create /tmp/$PRIVATE/snap.$SNAP_NAME/tmp 01777 root:root Ignore EEXIST since we
 	// want to reuse and we will open with O_NOFOLLOW, below.
 	if (mkdirat(base_dir_fd, "tmp", 01777) < 0 && errno != EEXIST) {
 		die("cannot create private tmp directory %s/tmp", base);
@@ -215,12 +216,14 @@ struct sc_mount_config {
  * Ensures all required mount points have been created
  */
 static void sc_create_mount_points(const char *scratch_dir,
-                                   const struct sc_mount *mounts)
+				   const struct sc_mount *mounts)
 {
 	char dst[PATH_MAX] = { 0 };
 	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
-	for (const struct sc_mount * mnt = mounts; mnt && mnt->path != NULL; mnt++) {
-		sc_must_snprintf(dst, sizeof(dst), "%s/%s", scratch_dir, mnt->path);
+	for (const struct sc_mount * mnt = mounts; mnt && mnt->path != NULL;
+	     mnt++) {
+		sc_must_snprintf(dst, sizeof(dst), "%s/%s", scratch_dir,
+				 mnt->path);
 		if (sc_nonfatal_mkpath(dst, 0755) < 0) {
 			die("cannot create mount point %s", dst);
 		}
@@ -240,8 +243,7 @@ static void sc_create_mount_points(const char *scratch_dir,
  * - All the target directories must exist
  * - All the source directories must exist, unless the mount is bi-directional
  */
-static void sc_do_mounts(const char *scratch_dir,
-                         const struct sc_mount *mounts)
+static void sc_do_mounts(const char *scratch_dir, const struct sc_mount *mounts)
 {
 	char dst[PATH_MAX] = { 0 };
 	// Bind mount certain directories from the host filesystem to the scratch
@@ -249,7 +251,8 @@ static void sc_do_mounts(const char *scratch_dir,
 	// of the peer group. This way the running application can alter any global
 	// state visible on the host and in other snaps. This can be restricted by
 	// disabling the "is_bidirectional" flag as can be seen below.
-	for (const struct sc_mount * mnt = mounts; mnt && mnt->path != NULL; mnt++) {
+	for (const struct sc_mount * mnt = mounts; mnt && mnt->path != NULL;
+	     mnt++) {
 
 		if (mnt->is_bidirectional) {
 			sc_identity old =
@@ -319,7 +322,7 @@ static void sc_initialize_ns_fstab(const char *snap_instance_name)
 		die("cannot open %s", info_path);
 	}
 	if (fchown(fd, 0, 0) < 0) {
-		die("cannot chown %s to root.root", info_path);
+		die("cannot chown %s to root:root", info_path);
 	}
 	// The stream now owns the file descriptor.
 	stream = fdopen(fd, "w");
@@ -351,8 +354,8 @@ static void sc_initialize_ns_fstab(const char *snap_instance_name)
  * not touch them.
  */
 static void sc_replicate_base_rootfs(const char *scratch_dir,
-                                     const char *rootfs_dir,
-                                     const struct sc_mount *root_mounts)
+				     const char *rootfs_dir,
+				     const struct sc_mount *root_mounts)
 {
 	// First of all, fix the root filesystem:
 	// - remove write permissions for group and others
@@ -371,13 +374,12 @@ static void sc_replicate_base_rootfs(const char *scratch_dir,
 	if (rootfs_fd < 0) {
 		die("cannot open directory \"%s\"", rootfs_dir);
 	}
-
 	// rootfs_fd is now managed by fdopendir() and should not be used after
 	DIR *rootfs SC_CLEANUP(sc_cleanup_closedir) = fdopendir(rootfs_fd);
 	if (rootfs == NULL) {
-		die("cannot open directory \"%s\" from file descriptor", rootfs_dir);
+		die("cannot open directory \"%s\" from file descriptor",
+		    rootfs_dir);
 	}
-
 	// Will create folders/links as 0:0
 	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
 
@@ -393,28 +395,30 @@ static void sc_replicate_base_rootfs(const char *scratch_dir,
 	while (true) {
 		errno = 0;
 		struct dirent *ent = readdir(rootfs);
-		if (ent == NULL) break;
+		if (ent == NULL)
+			break;
 
 		if (sc_streq(ent->d_name, ".") || sc_streq(ent->d_name, "..")) {
 			continue;
 		}
 
 		sc_must_snprintf(full_path, sizeof(full_path), "%s/%s",
-		                 scratch_dir, ent->d_name);
+				 scratch_dir, ent->d_name);
 		if (ent->d_type == DT_DIR) {
 			if (mkdir(full_path, 0755) < 0) {
-				die("cannot create directory \"%s\"", full_path);
+				die("cannot create directory \"%s\"",
+				    full_path);
 			}
-
 			// If the directory is listed in root_mounts skip it,
 			// as it will be created and mounted in
 			// sc_bootstrap_mount_namespace() later.
 			bool skip_dir = false;
-			const char *path_in_rootfs = full_path + scratch_dir_length;
-			for (const struct sc_mount *mnt = root_mounts;
+			const char *path_in_rootfs =
+			    full_path + scratch_dir_length;
+			for (const struct sc_mount * mnt = root_mounts;
 			     mnt->path != NULL; mnt++) {
-				if (sc_streq(path_in_rootfs, mnt->path) ||
-				    sc_streq(path_in_rootfs, mnt->altpath)) {
+				if (sc_streq(path_in_rootfs, mnt->path)
+				    || sc_streq(path_in_rootfs, mnt->altpath)) {
 					skip_dir = true;
 					break;
 				}
@@ -430,11 +434,13 @@ static void sc_replicate_base_rootfs(const char *scratch_dir,
 			char src_path[PATH_MAX];
 			sc_must_snprintf(src_path, sizeof(src_path), "%s/%s",
 					 rootfs_dir, ent->d_name);
-			sc_do_mount(src_path, full_path, NULL, MS_REC | MS_BIND, NULL);
+			sc_do_mount(src_path, full_path, NULL, MS_REC | MS_BIND,
+				    NULL);
 		} else if (ent->d_type == DT_LNK) {
 			char link_target[PATH_MAX + 1];
 			ssize_t len = readlinkat(rootfs_fd, ent->d_name,
-			                         link_target, sizeof(link_target) - 1);
+						 link_target,
+						 sizeof(link_target) - 1);
 			if (len < 0) {
 				die("cannot read symbolic link \"%s/%s\"",
 				    rootfs_dir, ent->d_name);
@@ -445,13 +451,15 @@ static void sc_replicate_base_rootfs(const char *scratch_dir,
 			// Both relative and absolute links will work out of the box, since
 			// we are going to do a pivot_root to scratch_dir.
 			if (symlink(link_target, full_path) < 0) {
-				die("cannot create symbolic link \"%s\"", full_path);
+				die("cannot create symbolic link \"%s\"",
+				    full_path);
 			}
 		} else if (ent->d_type == DT_REG) {
 			// Create an empty file which can be used as a mount point
 			int fd = open(full_path, O_CREAT | O_TRUNC, 0644);
 			if (fd < 0) {
-				die("cannot create mount point for file \"%s\"", full_path);
+				die("cannot create mount point for file \"%s\"",
+				    full_path);
 			}
 			close(fd);
 			char src_path[PATH_MAX];
@@ -459,8 +467,7 @@ static void sc_replicate_base_rootfs(const char *scratch_dir,
 					 rootfs_dir, ent->d_name);
 			sc_do_mount(src_path, full_path, NULL, MS_BIND, NULL);
 		} else {
-			die("unexpected directory entry \"%s\" of type %i encountered in \"%s\"",
-				ent->d_name, ent->d_type, rootfs_dir);
+			die("unexpected directory entry \"%s\" of type %i encountered in \"%s\"", ent->d_name, ent->d_type, rootfs_dir);
 		}
 	}
 
@@ -536,7 +543,8 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 		// Create a tmpfs on scratch_dir; we'll them mount all the root
 		// directories of the base snap onto it.
 		sc_do_mount("none", scratch_dir, "tmpfs", 0, NULL);
-		sc_replicate_base_rootfs(scratch_dir, config->rootfs_dir, config->mounts);
+		sc_replicate_base_rootfs(scratch_dir, config->rootfs_dir,
+					 config->mounts);
 	} else {
 		// Recursively bind mount desired root filesystem directory over the
 		// scratch directory. This puts the initial content into the scratch
@@ -545,8 +553,8 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 		//
 		// The mount is recursive because we need to accurately replicate the
 		// state of the root filesystem into the scratch directory.
-		sc_do_mount(config->rootfs_dir, scratch_dir, NULL, MS_REC | MS_BIND,
-		            NULL);
+		sc_do_mount(config->rootfs_dir, scratch_dir, NULL,
+			    MS_REC | MS_BIND, NULL);
 	}
 	// Make the scratch directory recursively slave. Nothing done there will be
 	// shared with the initial mount namespace. This effectively detaches us,
@@ -705,36 +713,41 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// already contains the correct view of the mounted snaps.
 	if (config->normal_mode) {
 		sc_must_snprintf(dst, sizeof dst, "%s/snap", scratch_dir);
-		sc_do_mount(SNAP_MOUNT_DIR, dst, NULL, MS_BIND | MS_REC, NULL);
+		sc_do_mount(sc_snap_mount_dir(NULL), dst, NULL,
+			    MS_BIND | MS_REC, NULL);
 		sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
 	}
 	// Ensure that hostfs exists and is group-owned by root. We may have (now
 	// or earlier) created the directory as the user who first ran a snap on a
 	// given system and the group identity of that user is visible on disk.
-	// This was LP:#1665004
-	struct stat sb;
-	if (stat(SC_HOSTFS_DIR, &sb) < 0) {
-		if (errno == ENOENT) {
-			// Create the hostfs directory if one is missing. This directory is a part
-			// of packaging now so perhaps this code can be removed later.
-			// Note: we use 0000 as permissions here, to avoid the risk that
-			// the user manages to fiddle with the newly created directory
-			// before we have the chance to chown it to root:root. We are
-			// setting the usual 0755 permissions just after the chown below.
-			if (mkdir(SC_HOSTFS_DIR, 0000) < 0) {
-				die("cannot perform operation: mkdir %s", SC_HOSTFS_DIR);
-			}
-			if (chown(SC_HOSTFS_DIR, 0, 0) < 0) {
-				die("cannot set root ownership on %s directory", SC_HOSTFS_DIR);
-			}
-			if (chmod(SC_HOSTFS_DIR, 0755) < 0) {
-				die("cannot set 0755 permissions on %s directory", SC_HOSTFS_DIR);
+	// This was LP:#1665004. We do this by trying to create the hostfs directory
+	// if one is missing. This directory is a part of packaging now so perhaps
+	// this code can be removed later. Note: we use 0000 as permissions here, to
+	// avoid the risk that the user manages to fiddle with the newly created
+	// directory before we have the chance to chown it to root:root. We are
+	// setting the usual 0755 permissions just after the chown below.
+	if (mkdir(SC_HOSTFS_DIR, 0000) < 0) {
+		if (errno == EEXIST) {
+			// The directory exists, verify its ownership.
+			struct stat sb;
+			if (stat(SC_HOSTFS_DIR, &sb) < 0) {
+				die("cannot stat %s", SC_HOSTFS_DIR);
+			} else if (sb.st_uid != 0 || sb.st_gid != 0) {
+				die("%s is not owned by root", SC_HOSTFS_DIR);
 			}
 		} else {
-			die("cannot stat %s", SC_HOSTFS_DIR);
+			die("cannot perform operation: mkdir %s",
+			    SC_HOSTFS_DIR);
+		}
+	} else {
+		if (chown(SC_HOSTFS_DIR, 0, 0) < 0) {
+			die("cannot set root ownership on %s directory",
+			    SC_HOSTFS_DIR);
+		}
+		if (chmod(SC_HOSTFS_DIR, 0755) < 0) {
+			die("cannot set 0755 permissions on %s directory",
+			    SC_HOSTFS_DIR);
 		}
-	} else if (sb.st_uid != 0 || sb.st_gid != 0) {
-		die("%s is not owned by root", SC_HOSTFS_DIR);
 	}
 	// Make the upcoming "put_old" directory for pivot_root private so that
 	// mount events don't propagate to any peer group. In practice pivot root
@@ -908,18 +921,18 @@ static struct sc_mount *sc_homedir_mounts(const struct sc_invocation *inv)
 	if (inv->num_homedirs == 0) {
 		return NULL;
 	}
-
 	// We add one element for the end-of-array indicator.
-	struct sc_mount *mounts = calloc(inv->num_homedirs + 1, sizeof(struct sc_mount));
+	struct sc_mount *mounts =
+	    calloc(inv->num_homedirs + 1, sizeof(struct sc_mount));
 	if (mounts == NULL) {
 		die("cannot allocate mount data for homedirs");
 	}
-
 	// Copy inv->homedirs to the mount structures
 	for (int i = 0; i < inv->num_homedirs; i++) {
 		debug("Adding homedir: %s", inv->homedirs[i]);
 		mounts[i].path = sc_strdup(inv->homedirs[i]);
-		mounts[i].is_bidirectional = true;
+		// Note that we are not setting bidirectional flag, so anything mounted
+		// here will not propagate to the host.
 	}
 	return mounts;
 }
@@ -930,18 +943,17 @@ static void sc_free_dynamic_mounts(struct sc_mount *mounts)
 	if (mounts == NULL) {
 		return;
 	}
-
 	// Cleanup allocated resources by each of the mount
 	// structures. The array will be terminated by a single zeroed
 	// entry.
 	for (int i = 0; mounts[i].path != NULL; i++) {
-		free((void*)mounts[i].path);
+		free((void *)mounts[i].path);
 	}
 	free(mounts);
 }
 
 void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
-			  const sc_invocation * inv, const gid_t real_gid,
+			  const sc_invocation *inv, const gid_t real_gid,
 			  const gid_t saved_gid)
 {
 	// Classify the current distribution, as claimed by /etc/os-release.
@@ -951,32 +963,32 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 	if (inv->is_normal_mode) {
 		// In normal mode we use the base snap as / and set up several bind mounts.
 		static const struct sc_mount mounts[] = {
-			{"/dev"},	// because it contains devices on host OS
-			{"/etc"},	// because that's where /etc/resolv.conf lives, perhaps a bad idea
-			{"/home"},	// to support /home/*/snap and home interface
-			{"/root"},	// because that is $HOME for services
-			{"/proc"},	// fundamental filesystem
-			{"/sys"},	// fundamental filesystem
-			{"/tmp"},	// to get writable tmp
-			{"/var/snap"},	// to get access to global snap data
-			{"/var/lib/snapd"},	// to get access to snapd state and seccomp profiles
-			{"/var/tmp"},	// to get access to the other temporary directory
-			{"/run"},	// to get /run with sockets and what not
-			{"/lib/modules",.is_optional = true},	// access to the modules of the running kernel
-			{"/lib/firmware",.is_optional = true},	// access to the firmware of the running kernel
-			{"/usr/src"},	// FIXME: move to SecurityMounts in system-trace interface
-			{"/var/log"},	// FIXME: move to SecurityMounts in log-observe interface
+			{.path = "/dev"},	// because it contains devices on host OS
+			{.path = "/etc"},	// because that's where /etc/resolv.conf lives, perhaps a bad idea
+			{.path = "/home"},	// to support /home/*/snap and home interface
+			{.path = "/root"},	// because that is $HOME for services
+			{.path = "/proc"},	// fundamental filesystem
+			{.path = "/sys"},	// fundamental filesystem
+			{.path = "/tmp"},	// to get writable tmp
+			{.path = "/var/snap"},	// to get access to global snap data
+			{.path = "/var/lib/snapd"},	// to get access to snapd state and seccomp profiles
+			{.path = "/var/tmp"},	// to get access to the other temporary directory
+			{.path = "/run"},	// to get /run with sockets and what not
+			{.path = "/lib/modules",.is_optional = true},	// access to the modules of the running kernel
+			{.path = "/lib/firmware",.is_optional = true},	// access to the firmware of the running kernel
+			{.path = "/usr/src"},	// FIXME: move to SecurityMounts in system-trace interface
+			{.path = "/var/log"},	// FIXME: move to SecurityMounts in log-observe interface
 #ifdef MERGED_USR
-			{"/run/media", true, "/media"},	// access to the users removable devices
+			{.path = "/run/media",.is_bidirectional = true,.altpath = "/media"},	// access to the users removable devices
 #else
-			{"/media", true},	// access to the users removable devices
+			{.path = "/media",.is_bidirectional = true},	// access to the users removable devices
 #endif				// MERGED_USR
-			{"/run/netns", true},	// access to the 'ip netns' network namespaces
+			{.path = "/run/netns",.is_bidirectional = true},	// access to the 'ip netns' network namespaces
 			// The /mnt directory is optional in base snaps to ensure backwards
 			// compatibility with the first version of base snaps that was
 			// released.
-			{"/mnt",.is_optional = true},	// to support the removable-media interface
-			{"/var/lib/extrausers",.is_optional = true},	// access to UID/GID of extrausers (if available)
+			{.path = "/mnt",.is_optional = true},	// to support the removable-media interface
+			{.path = "/var/lib/extrausers",.is_optional = true},	// access to UID/GID of extrausers (if available)
 			{},
 		};
 		struct sc_mount_config normal_config = {
@@ -998,8 +1010,8 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 		// In legacy mode we don't pivot to a base snap's rootfs and instead
 		// just arrange bi-directional mount propagation for two directories.
 		static const struct sc_mount mounts[] = {
-			{"/media", true},
-			{"/run/netns", true},
+			{.path = "/media",.is_bidirectional = true},
+			{.path = "/run/netns",.is_bidirectional = true},
 			{},
 		};
 		struct sc_mount_config legacy_config = {
@@ -1050,16 +1062,16 @@ static bool is_mounted_with_shared_option(const char *dir)
 void sc_ensure_shared_snap_mount(void)
 {
 	if (!is_mounted_with_shared_option("/")
-	    && !is_mounted_with_shared_option(SNAP_MOUNT_DIR)) {
+	    && !is_mounted_with_shared_option(sc_snap_mount_dir(NULL))) {
 		// TODO: We could be more aggressive and refuse to function but since
 		// we have no data on actual environments that happen to limp along in
 		// this configuration let's not do that yet.  This code should be
 		// removed once we have a measurement and feedback mechanism that lets
 		// us decide based on measurable data.
-		sc_do_mount(SNAP_MOUNT_DIR, SNAP_MOUNT_DIR, "none",
-			    MS_BIND | MS_REC, NULL);
-		sc_do_mount("none", SNAP_MOUNT_DIR, NULL, MS_SHARED | MS_REC,
-			    NULL);
+		sc_do_mount(sc_snap_mount_dir(NULL), sc_snap_mount_dir(NULL),
+			    "none", MS_BIND | MS_REC, NULL);
+		sc_do_mount("none", sc_snap_mount_dir(NULL), NULL,
+			    MS_SHARED | MS_REC, NULL);
 	}
 }
 
@@ -1089,7 +1101,7 @@ void sc_setup_user_mounts(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 
 void sc_ensure_snap_dir_shared_mounts(void)
 {
-	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	const char *dirs[] = { sc_snap_mount_dir(NULL), "/var/snap", NULL };
 	for (int i = 0; dirs[i] != NULL; i++) {
 		const char *dir = dirs[i];
 		if (!is_mounted_with_shared_option(dir)) {
@@ -1116,16 +1128,17 @@ void sc_setup_parallel_instance_classic_mounts(const char *snap_name,
 	char src[PATH_MAX] = { 0 };
 	char dst[PATH_MAX] = { 0 };
 
-	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	const char *dirs[] = { sc_snap_mount_dir(NULL), "/var/snap", NULL };
 	for (int i = 0; dirs[i] != NULL; i++) {
 		const char *dir = dirs[i];
 		sc_do_mount("none", dir, NULL, MS_REC | MS_SLAVE, NULL);
 	}
 
 	/* Mount SNAP_MOUNT_DIR/<snap>_<key> on SNAP_MOUNT_DIR/<snap> */
-	sc_must_snprintf(src, sizeof src, "%s/%s", SNAP_MOUNT_DIR,
+	sc_must_snprintf(src, sizeof src, "%s/%s", sc_snap_mount_dir(NULL),
 			 snap_instance_name);
-	sc_must_snprintf(dst, sizeof dst, "%s/%s", SNAP_MOUNT_DIR, snap_name);
+	sc_must_snprintf(dst, sizeof dst, "%s/%s", sc_snap_mount_dir(NULL),
+			 snap_name);
 	sc_do_mount(src, dst, "none", MS_BIND | MS_REC, NULL);
 
 	/* Mount /var/snap/<snap>_<key> on /var/snap/<snap> */
