@@ -21,7 +21,6 @@ package main
 
 import (
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"strings"
@@ -66,7 +65,10 @@ var (
 	sysFchdir     = syscall.Fchdir
 	sysLstat      = syscall.Lstat
 
-	ioutilReadDir = ioutil.ReadDir
+	sysGetuid = sys.Getuid
+	sysGetgid = sys.Getgid
+
+	osReadDir = os.ReadDir
 )
 
 // ReadOnlyFsError is an error encapsulating encountered EROFS.
@@ -108,7 +110,7 @@ func OpenPath(path string) (int, error) {
 		if !strings.HasSuffix(iter.CurrentName(), "/") {
 			openFlags &^= syscall.O_DIRECTORY
 		}
-		fd, err = sysOpenat(fd, iter.CurrentCleanName(), openFlags, 0)
+		fd, err = sysOpenat(fd, iter.CurrentNameNoSlash(), openFlags, 0)
 		if err != nil {
 			return -1, err
 		}
@@ -127,6 +129,23 @@ func OpenPath(path string) (int, error) {
 	return fd, nil
 }
 
+// syscallMode returns the syscall-specific mode bits from Go's portable mode bits.
+// This is a copy of the same helper in Go's os package.
+func syscallMode(i os.FileMode) (o uint32) {
+	o |= uint32(i.Perm())
+	if i&os.ModeSetuid != 0 {
+		o |= syscall.S_ISUID
+	}
+	if i&os.ModeSetgid != 0 {
+		o |= syscall.S_ISGID
+	}
+	if i&os.ModeSticky != 0 {
+		o |= syscall.S_ISVTX
+	}
+	// No mapping for Go's ModeTemporary (plan9 only).
+	return o
+}
+
 // MkPrefix creates all the missing directories in a given base path and
 // returns the file descriptor to the leaf directory as well as the restricted
 // flag. This function is a base for secure variants of mkdir, touch and
@@ -156,7 +175,7 @@ func MkPrefix(base string, perm os.FileMode, uid sys.UserID, gid sys.GroupID, rs
 		// Keep closing the previous descriptor as we go, so that we have the
 		// last one handy from the MkDir below.
 		defer sysClose(fd)
-		fd, err = MkDir(fd, iter.CurrentBase(), iter.CurrentCleanName(), perm, uid, gid, rs)
+		fd, err = MkDir(fd, iter.CurrentBaseNoSlash(), iter.CurrentNameNoSlash(), perm, uid, gid, rs)
 		if err != nil {
 			return -1, err
 		}
@@ -179,7 +198,7 @@ func MkDir(dirFd int, dirName string, name string, perm os.FileMode, uid sys.Use
 	made := true
 	const openFlags = syscall.O_NOFOLLOW | syscall.O_CLOEXEC | syscall.O_DIRECTORY
 
-	if err := sysMkdirat(dirFd, name, uint32(perm.Perm())); err != nil {
+	if err := sysMkdirat(dirFd, name, syscallMode(perm)); err != nil {
 		switch err {
 		case syscall.EEXIST:
 			made = false
@@ -202,7 +221,7 @@ func MkDir(dirFd int, dirName string, name string, perm os.FileMode, uid sys.Use
 			// Close the FD we opened if we fail here since the caller will get
 			// an error and won't assume responsibility for the FD.
 			sysClose(newFd)
-			return -1, fmt.Errorf("cannot chown directory %q to %d.%d: %v", filepath.Join(dirName, name), uid, gid, err)
+			return -1, fmt.Errorf("cannot chown directory %q to %d:%d: %v", filepath.Join(dirName, name), uid, gid, err)
 		}
 		// As soon as we find a place that is safe to write we can switch off
 		// the restricted mode (and thus any subsequent checks). This is
@@ -238,7 +257,7 @@ func MkFile(dirFd int, dirName string, name string, perm os.FileMode, uid sys.Us
 	// we know if we need to chown it) but fall back to just opening an
 	// existing one.
 
-	newFd, err := sysOpenat(dirFd, name, openFlags|syscall.O_CREAT|syscall.O_EXCL, uint32(perm.Perm()))
+	newFd, err := sysOpenat(dirFd, name, openFlags|syscall.O_CREAT|syscall.O_EXCL, syscallMode(perm))
 	if err != nil {
 		switch err {
 		case syscall.EEXIST:
@@ -262,7 +281,7 @@ func MkFile(dirFd int, dirName string, name string, perm os.FileMode, uid sys.Us
 	if made {
 		// Chown the file if we made it.
 		if err := sysFchown(newFd, uid, gid); err != nil {
-			return fmt.Errorf("cannot chown file %q to %d.%d: %v", filepath.Join(dirName, name), uid, gid, err)
+			return fmt.Errorf("cannot chown file %q to %d:%d: %v", filepath.Join(dirName, name), uid, gid, err)
 		}
 	}
 
@@ -322,6 +341,119 @@ func MkSymlink(dirFd int, dirName string, name string, oldname string, rs *Restr
 	return nil
 }
 
+// MkdirAllWithin is the secure variant of os.MkdirAll that creates all the missing directories of the given path within the
+// given existing parent directory.
+//
+// Unlike os.MkdirAll this implementation does not follow any symbolic
+// links. At all times the new directory segment is created using mkdirat(2)
+// while holding an open file descriptor to the parent directory.
+//
+// The only handled error is mkdirat(2) that fails with EEXIST. All other
+// errors are fatal but there is no attempt to undo anything that was created.
+//
+// The uid and gid are used for the fchown(2) system call which is performed
+// after each segment is created and opened. The special value -1 may be used
+// to request that ownership is not changed.
+func MkdirAllWithin(path, parent string, perm os.FileMode, uid sys.UserID, gid sys.GroupID, rs *Restrictions) error {
+	path = filepath.Clean(path)
+	parent = filepath.Clean(parent)
+	if !filepath.IsAbs(path) {
+		return fmt.Errorf("cannot use relative path %q", path)
+	}
+	if !filepath.IsAbs(parent) {
+		return fmt.Errorf("cannot use relative parent path %q", parent)
+	}
+	isParent := func(path, parent string) bool {
+		if path == parent {
+			return false
+		}
+		if parent == "/" {
+			return true
+		}
+		return strings.HasPrefix(path, parent+string(filepath.Separator))
+	}
+	if !isParent(path, parent) {
+		return fmt.Errorf("path %q is not a parent of %q", parent, path)
+	}
+
+	// Check if we need to do anything
+	fi, err := osLstat(path)
+	if err == nil {
+		if !fi.Mode().IsDir() {
+			return fmt.Errorf("cannot create directory %q: existing file in the way", path)
+		}
+		return nil
+	} else if !os.IsNotExist(err) {
+		return fmt.Errorf("cannot inspect path %q: %v", path, err)
+	}
+
+	// Check that the parent path exists
+	fi, err = osLstat(parent)
+	if err == nil {
+		if !fi.Mode().IsDir() {
+			return fmt.Errorf("cannot use parent path %q: not a directory", parent)
+		}
+	} else if os.IsNotExist(err) {
+		return fmt.Errorf("parent directory %q does not exist", parent)
+	} else {
+		return fmt.Errorf("cannot inspect parent path %q: %v", parent, err)
+	}
+
+	iter, err := strutil.NewPathIterator(path)
+	if err != nil {
+		return fmt.Errorf("cannot iterate over path %q: %v", path, err)
+	}
+	// Advance the iterator to the parent. Finding the parent is
+	// guaranteed by the earlier check isParent.
+	for iter.Next() {
+		if iter.CurrentPathNoSlash() == parent {
+			break
+		}
+	}
+	// Advance the iterator to the first missing directory
+	for iter.Next() {
+		if iter.CurrentPathNoSlash() == path {
+			// Already confirmed path does not exist
+			break
+		}
+		fi, err = osLstat(iter.CurrentPathNoSlash())
+		if err == nil {
+			if !fi.Mode().IsDir() {
+				return fmt.Errorf("cannot create directory %q: existing file in the way", iter.CurrentPathNoSlash())
+			}
+			continue
+		}
+		if !os.IsNotExist(err) {
+			return fmt.Errorf("cannot inspect path %q: %v", iter.CurrentPathNoSlash(), err)
+		}
+		break
+	}
+
+	// Create the first missing directory. From this point onward all file descriptors are kept open
+	// until all missing directories have been created or failure, and then closed in reverse order.
+	const openFlags = syscall.O_NOFOLLOW | syscall.O_CLOEXEC | syscall.O_DIRECTORY
+	fd, err := sysOpen(iter.CurrentBaseNoSlash(), openFlags, 0)
+	if err != nil {
+		return fmt.Errorf("cannot open directory %q: %v", iter.CurrentBaseNoSlash(), err)
+	}
+	defer sysClose(fd)
+	fd, err = MkDir(fd, iter.CurrentBaseNoSlash(), iter.CurrentNameNoSlash(), perm, uid, gid, rs)
+	if err != nil {
+		return err
+	}
+	defer sysClose(fd)
+
+	// Create the remaining missing directories
+	for iter.Next() {
+		fd, err = MkDir(fd, iter.CurrentBaseNoSlash(), iter.CurrentNameNoSlash(), perm, uid, gid, rs)
+		if err != nil {
+			return err
+		}
+		defer sysClose(fd)
+	}
+	return nil
+}
+
 // MkdirAll is the secure variant of os.MkdirAll.
 //
 // Unlike the regular version this implementation does not follow any symbolic
@@ -501,7 +633,7 @@ func planWritableMimic(dir, neededBy string) ([]*Change, error) {
 		},
 	})
 	// Iterate over the items in the original directory (nothing is mounted _yet_).
-	entries, err := ioutilReadDir(dir)
+	entries, err := osReadDir(dir)
 	if err != nil {
 		return nil, err
 	}
@@ -513,7 +645,7 @@ func planWritableMimic(dir, neededBy string) ([]*Change, error) {
 		// Bind mount each element from the safe-keeping directory into the
 		// tmpfs. Our Change.Perform() engine can create the missing
 		// directories automatically so we don't bother creating those.
-		m := fi.Mode()
+		m := fi.Type()
 		switch {
 		case m.IsDir():
 			ch.Entry.Options = []string{"rbind"}
@@ -552,11 +684,11 @@ type FatalError struct {
 // Certain assumptions are made about the plan, it must closely resemble that
 // created by planWritableMimic, in particular the sequence must look like this:
 //
-// - bind a directory aside into safekeeping location
-// - cover the original with tmpfs
-// - bind mount something from safekeeping location to an empty file or
-//   directory in the tmpfs; this step can repeat any number of times
-// - unbind the safekeeping location
+//   - bind a directory aside into safekeeping location
+//   - cover the original with tmpfs
+//   - bind mount something from safekeeping location to an empty file or
+//     directory in the tmpfs; this step can repeat any number of times
+//   - unbind the safekeeping location
 //
 // Apart from merely executing the plan a fake plan is returned for undo. The
 // undo plan skips the following elements as compared to the original plan:
