@@ -25,7 +25,6 @@ import (
 	"crypto/rand"
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"net/http"
 	"os"
 	"path/filepath"
@@ -39,12 +38,15 @@ import (
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/daemon"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/assertstate/assertstatetest"
 	"github.com/snapcore/snapd/overlord/snapstate"
+	"github.com/snapcore/snapd/overlord/snapstate/sequence"
 	"github.com/snapcore/snapd/overlord/snapstate/snapstatetest"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/sandbox"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/testutil"
@@ -132,6 +134,33 @@ func (s *sideloadSuite) TestSideloadSnapDevMode(c *check.C) {
 	c.Check(chgSummary, check.Equals, `Install "local" snap from file "x"`)
 }
 
+func (s *sideloadSuite) TestSideloadSnapQuotaGroup(c *check.C) {
+	body := "" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"snap\"; filename=\"x\"\r\n" +
+		"\r\n" +
+		"xyzzy\r\n" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"devmode\"\r\n" +
+		"\r\n" +
+		"true\r\n" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"quota-group\"\r\n" +
+		"\r\n" +
+		"foo\r\n" +
+		"----hello--\r\n"
+	head := map[string]string{"Content-Type": "multipart/thing; boundary=--hello--"}
+	// try a multipart/form-data upload
+	flags := snapstate.Flags{
+		RemoveSnapPath: true,
+		DevMode:        true,
+		Transaction:    client.TransactionPerSnap,
+		QuotaGroupName: "foo",
+	}
+	chgSummary, _ := s.sideloadCheck(c, body, head, "local", flags)
+	c.Check(chgSummary, check.Equals, `Install "local" snap from file "x"`)
+}
+
 func (s *sideloadSuite) TestSideloadSnapJailMode(c *check.C) {
 	body := "" +
 		"----hello--\r\n" +
@@ -175,16 +204,20 @@ func (s *sideloadSuite) sideloadCheck(c *check.C, content string, head map[strin
 		return &snap.Info{SuggestedName: mockedName}, nil
 	})()
 
-	defer daemon.MockSnapstateInstall(func(ctx context.Context, s *state.State, name string, opts *snapstate.RevisionOptions, userID int, flags snapstate.Flags) (*state.TaskSet, error) {
-		// NOTE: ubuntu-core is not installed in developer mode
-		c.Check(flags, check.Equals, snapstate.Flags{})
-		installQueue = append(installQueue, name)
+	defer daemon.MockSnapstateInstallWithGoal(func(ctx context.Context, st *state.State, g snapstate.InstallGoal, opts snapstate.Options) ([]*snap.Info, []*state.TaskSet, error) {
+		goal, ok := g.(*storeInstallGoalRecorder)
+		c.Assert(ok, check.Equals, true, check.Commentf("unexpected InstallGoal type %T", g))
+		c.Assert(goal.snaps, check.HasLen, 1)
 
-		t := s.NewTask("fake-install-snap", "Doing a fake install")
-		return state.NewTaskSet(t), nil
+		// NOTE: ubuntu-core is not installed in developer mode
+		c.Check(opts.Flags, check.Equals, snapstate.Flags{})
+		installQueue = append(installQueue, goal.snaps[0].InstanceName)
+
+		t := st.NewTask("fake-install-snap", "Doing a fake install")
+		return []*snap.Info{{}}, []*state.TaskSet{state.NewTaskSet(t)}, nil
 	})()
 
-	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags) (*state.TaskSet, *snap.Info, error) {
+	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags, prqt snapstate.PrereqTracker) (*state.TaskSet, *snap.Info, error) {
 		c.Check(flags, check.DeepEquals, expectedFlags)
 
 		c.Check(path, testutil.FileEquals, "xyzzy")
@@ -231,7 +264,239 @@ func (s *sideloadSuite) sideloadCheck(c *check.C, content string, head map[strin
 	err = chg.Get("api-data", &apiData)
 	c.Assert(err, check.IsNil)
 	c.Check(apiData, check.DeepEquals, map[string]interface{}{
-		"snap-name": expectedInstanceName,
+		"snap-name":  expectedInstanceName,
+		"snap-names": []interface{}{expectedInstanceName},
+	})
+
+	summary = chg.Summary()
+	err = chg.Get("system-restart-immediate", &systemRestartImmediate)
+	if err != nil && !errors.Is(err, state.ErrNoState) {
+		c.Error(err)
+	}
+	return summary, systemRestartImmediate
+}
+
+const sideLoadComponentBody = "" +
+	"----hello--\r\n" +
+	"Content-Disposition: form-data; name=\"snap\"; filename=\"x\"\r\n" +
+	"\r\n" +
+	"xyzzy\r\n" +
+	"----hello--\r\n" +
+	"Content-Disposition: form-data; name=\"dangerous\"\r\n" +
+	"\r\n" +
+	"true\r\n" +
+	"----hello--\r\n" +
+	"Content-Disposition: form-data; name=\"snap-path\"\r\n" +
+	"\r\n" +
+	"a/b/local+localcomp.comp\r\n" +
+	"----hello--\r\n"
+
+func (s *sideloadSuite) TestSideloadComponent(c *check.C) {
+	// try a multipart/form-data upload
+	body := sideLoadComponentBody
+	head := map[string]string{"Content-Type": "multipart/thing; boundary=--hello--"}
+	flags := snapstate.Flags{RemoveSnapPath: true, Transaction: client.TransactionPerSnap}
+	csi := snap.NewComponentSideInfo(naming.NewComponentRef("local", "comp"), snap.Revision{})
+
+	chgSummary, systemRestartImmediate := s.sideloadComponentCheck(c, body, head, "local", flags, csi)
+	c.Check(chgSummary, check.Equals, `Install "comp" component for "local" snap from file "a/b/local+localcomp.comp"`)
+	c.Check(systemRestartImmediate, check.Equals, false)
+}
+
+func (s *sideloadSuite) TestSideloadComponentInstanceName(c *check.C) {
+	// try a multipart/form-data upload
+	body := sideLoadComponentBody +
+		"Content-Disposition: form-data; name=\"name\"\r\n" +
+		"\r\n" +
+		"local_instance\r\n" +
+		"----hello--\r\n"
+	head := map[string]string{"Content-Type": "multipart/thing; boundary=--hello--"}
+	flags := snapstate.Flags{RemoveSnapPath: true, Transaction: client.TransactionPerSnap}
+	csi := snap.NewComponentSideInfo(naming.NewComponentRef("local", "comp"), snap.Revision{})
+
+	chgSummary, systemRestartImmediate := s.sideloadComponentCheck(c, body, head, "local_instance", flags, csi)
+	c.Check(chgSummary, check.Equals, `Install "comp" component for "local_instance" snap from file "a/b/local+localcomp.comp"`)
+	c.Check(systemRestartImmediate, check.Equals, false)
+}
+
+func (s *sideloadSuite) TestSideloadComponentNoDangerousFlag(c *check.C) {
+	logbuf, r := logger.MockLogger()
+	defer r()
+	body := "----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"snap\"; filename=\"x\"\r\n" +
+		"\r\n" +
+		"xyzzy\r\n" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"snap-path\"\r\n" +
+		"\r\n" +
+		"a/b/local+localcomp.comp\r\n" +
+		"----hello--\r\n"
+	d := s.daemonWithFakeSnapManager(c)
+	s.markSeeded(d)
+
+	defer daemon.MockUnsafeReadSnapInfo(func(path string) (*snap.Info, error) {
+		return nil, daemon.BadRequest("mocking error to force reading as component")
+	})()
+
+	req, err := http.NewRequest("POST", "/v2/snaps", bytes.NewBufferString(body))
+	c.Assert(err, check.IsNil)
+	req.Header.Set("Content-Type", "multipart/thing; boundary=--hello--")
+
+	rspe := s.errorReq(c, req, nil)
+	c.Check(logbuf.String(), testutil.Contains,
+		"cannot sideload as a component: only unasserted installation of local component with --dangerous is supported at the moment")
+	c.Check(rspe.Message, check.Equals,
+		`cannot find signatures with metadata for snap "a/b/local+localcomp.comp"`)
+}
+
+func (s *sideloadSuite) TestSideloadComponentForNotInstalledSnap(c *check.C) {
+	logbuf, r := logger.MockLogger()
+	defer r()
+	body := "----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"snap\"; filename=\"x\"\r\n" +
+		"\r\n" +
+		"xyzzy\r\n" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"dangerous\"\r\n" +
+		"\r\n" +
+		"true\r\n" +
+		"----hello--\r\n" +
+		"Content-Disposition: form-data; name=\"snap-path\"\r\n" +
+		"\r\n" +
+		"a/b/local+localcomp.comp\r\n" +
+		"----hello--\r\n"
+	d := s.daemonWithFakeSnapManager(c)
+	s.markSeeded(d)
+
+	defer daemon.MockUnsafeReadSnapInfo(func(path string) (*snap.Info, error) {
+		return nil, daemon.BadRequest("mocking error to force reading as component")
+	})()
+	defer daemon.MockReadComponentInfoFromCont(func(tempPath string, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+		return &snap.ComponentInfo{
+			Component: naming.NewComponentRef("local", "comp"),
+			Type:      snap.TestComponent,
+			Version:   "1.0",
+		}, nil
+	})()
+
+	st := s.d.Overlord().State()
+	st.Lock()
+	ssi := &snap.SideInfo{RealName: "other", Revision: snap.R(1),
+		SnapID: "some-other-snap-id"}
+	snapstate.Set(st, "other", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromRevisionSideInfos(
+			[]*sequence.RevisionSideState{
+				sequence.NewRevisionSideState(ssi, nil)}),
+		Current: snap.R(1),
+	})
+	st.Unlock()
+
+	req, err := http.NewRequest("POST", "/v2/snaps", bytes.NewBufferString(body))
+	c.Assert(err, check.IsNil)
+	req.Header.Set("Content-Type", "multipart/thing; boundary=--hello--")
+
+	rspe := s.errorReq(c, req, nil)
+	c.Check(logbuf.String(), testutil.Contains,
+		`cannot sideload as a snap: cannot read snap file: mocking error to force reading as component`)
+	c.Check(logbuf.String(), testutil.Contains,
+		`cannot sideload as a component: snap owning "local+comp" not installed`)
+	c.Check(rspe.Message, check.Equals, `snap owning "local+comp" not installed`)
+	c.Check(rspe.Kind, check.Equals, client.ErrorKindSnapNotInstalled)
+}
+
+func (s *sideloadSuite) sideloadComponentCheck(c *check.C, content string,
+	head map[string]string, expectedInstanceName string, expectedFlags snapstate.Flags,
+	expectedCompSideInfo *snap.ComponentSideInfo) (
+	summary string, systemRestartImmediate bool) {
+
+	d := s.daemonWithFakeSnapManager(c)
+	s.markSeeded(d)
+
+	st := s.d.Overlord().State()
+
+	st.Lock()
+	defer st.Unlock()
+	ssi := &snap.SideInfo{RealName: "local", Revision: snap.R(1),
+		SnapID: "some-snap-id"}
+	snapstate.Set(st, expectedInstanceName, &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromRevisionSideInfos(
+			[]*sequence.RevisionSideState{
+				sequence.NewRevisionSideState(ssi, nil)}),
+		Current: snap.R(1),
+	})
+	st.Unlock()
+
+	soon := 0
+	var origEnsureStateSoon func(*state.State)
+	origEnsureStateSoon, restore := daemon.MockEnsureStateSoon(func(st *state.State) {
+		soon++
+		origEnsureStateSoon(st)
+	})
+	defer restore()
+
+	c.Assert(expectedInstanceName != "", check.Equals, true, check.Commentf("expected instance name must be set"))
+
+	// setup done
+	installQueue := []string{}
+	defer daemon.MockUnsafeReadSnapInfo(func(path string) (*snap.Info, error) {
+		return nil, daemon.BadRequest("mocking error to force reading as component")
+	})()
+
+	defer daemon.MockReadComponentInfoFromCont(func(tempPath string, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+		return snap.NewComponentInfo(
+			expectedCompSideInfo.Component,
+			snap.TestComponent,
+			"1.0", "", "", "", nil,
+		), nil
+	})()
+
+	defer daemon.MockSnapstateInstallComponentPath(func(st *state.State, csi *snap.ComponentSideInfo, info *snap.Info,
+		path string, flags snapstate.Flags) (*state.TaskSet, error) {
+		c.Check(csi, check.DeepEquals, expectedCompSideInfo)
+		c.Check(flags, check.DeepEquals, expectedFlags)
+		c.Check(path, testutil.FileEquals, "xyzzy")
+
+		installQueue = append(installQueue, csi.Component.String()+"::"+path)
+		t := st.NewTask("fake-install-component", "Doing a fake install")
+		return state.NewTaskSet(t), nil
+	})()
+
+	buf := bytes.NewBufferString(content)
+	req, err := http.NewRequest("POST", "/v2/snaps", buf)
+	c.Assert(err, check.IsNil)
+	for k, v := range head {
+		req.Header.Set(k, v)
+	}
+
+	rsp := s.asyncReq(c, req, nil)
+	n := 1
+	c.Assert(installQueue, check.HasLen, n)
+	c.Check(installQueue[n-1], check.Matches, "local\\+comp::.*/"+regexp.QuoteMeta(dirs.LocalInstallBlobTempPrefix)+".*")
+
+	st.Lock()
+	chg := st.Change(rsp.Change)
+	c.Assert(chg, check.NotNil)
+	c.Check(soon, check.Equals, 1)
+	c.Assert(chg.Tasks(), check.HasLen, n)
+
+	st.Unlock()
+	s.waitTrivialChange(c, chg)
+	st.Lock()
+
+	c.Check(chg.Kind(), check.Equals, "install-component")
+	var names []string
+	err = chg.Get("snap-names", &names)
+
+	c.Assert(err, check.IsNil)
+	c.Check(names, check.DeepEquals, []string{expectedInstanceName})
+	var apiData map[string]interface{}
+	err = chg.Get("api-data", &apiData)
+	c.Assert(err, check.IsNil)
+	c.Check(apiData, check.DeepEquals, map[string]interface{}{
+		"components": map[string]interface{}{
+			expectedInstanceName: []interface{}{expectedCompSideInfo.Component.ComponentName}},
 	})
 
 	summary = chg.Summary()
@@ -301,7 +566,7 @@ func (s *sideloadSuite) TestLocalInstallSnapDeriveSideInfo(c *check.C) {
 version: 1`, nil)
 	digest, size, err := asserts.SnapFileSHA3_384(fooSnap)
 	c.Assert(err, check.IsNil)
-	fooSnapBytes, err := ioutil.ReadFile(fooSnap)
+	fooSnapBytes, err := os.ReadFile(fooSnap)
 	c.Assert(err, check.IsNil)
 
 	dev1Acct := assertstest.NewAccount(s.StoreSigning, "devel1", nil, "")
@@ -340,7 +605,7 @@ version: 1`, nil)
 	c.Assert(err, check.IsNil)
 	req.Header.Set("Content-Type", "multipart/thing; boundary=--hello--")
 
-	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags) (*state.TaskSet, *snap.Info, error) {
+	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags, prqt snapstate.PrereqTracker) (*state.TaskSet, *snap.Info, error) {
 		c.Check(flags, check.Equals, snapstate.Flags{RemoveSnapPath: true, Transaction: client.TransactionPerSnap})
 		c.Check(si, check.DeepEquals, &snap.SideInfo{
 			RealName: "foo",
@@ -366,7 +631,8 @@ version: 1`, nil)
 	err = chg.Get("api-data", &apiData)
 	c.Assert(err, check.IsNil)
 	c.Check(apiData, check.DeepEquals, map[string]interface{}{
-		"snap-name": "foo",
+		"snap-name":  "foo",
+		"snap-names": []interface{}{"foo"},
 	})
 }
 
@@ -434,7 +700,7 @@ func (s *sideloadSuite) TestSideloadSnapChangeConflict(c *check.C) {
 		return &snap.Info{SuggestedName: "foo"}, nil
 	})()
 
-	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags) (*state.TaskSet, *snap.Info, error) {
+	defer daemon.MockSnapstateInstallPath(func(s *state.State, si *snap.SideInfo, path, name, channel string, flags snapstate.Flags, prqt snapstate.PrereqTracker) (*state.TaskSet, *snap.Info, error) {
 		return nil, nil, &snapstate.ChangeConflictError{Snap: "foo"}
 	})()
 
@@ -549,7 +815,7 @@ func (s *sideloadSuite) TestFormdataIsWrittenToCorrectTmpLocation(c *check.C) {
 	chgSummary, _ := s.sideloadCheck(c, sideLoadBodyWithoutDevMode, head, "local", snapstate.Flags{RemoveSnapPath: true, Transaction: client.TransactionPerSnap})
 	c.Check(chgSummary, check.Equals, `Install "local" snap from file "a/b/local.snap"`)
 
-	files, err := ioutil.ReadDir(tmpDir)
+	files, err := os.ReadDir(tmpDir)
 	c.Assert(err, check.IsNil)
 	c.Assert(files, check.HasLen, 0)
 
@@ -677,14 +943,13 @@ func (s *sideloadSuite) TestSideloadManySnaps(c *check.C) {
 	s.markSeeded(d)
 	expectedFlags := &snapstate.Flags{RemoveSnapPath: true, DevMode: true, Transaction: client.TransactionAllSnaps}
 
-	restore := daemon.MockSnapstateInstallPathMany(func(_ context.Context, s *state.State, infos []*snap.SideInfo, paths []string, userID int, flags *snapstate.Flags) ([]*state.TaskSet, error) {
+	restore := daemon.MockSnapstateInstallPathMany(func(_ context.Context, s *state.State, infos []*snap.SideInfo, tmpPaths []string, userID int, flags *snapstate.Flags) ([]*state.TaskSet, error) {
 		c.Check(flags, check.DeepEquals, expectedFlags)
 		c.Check(userID, check.Not(check.Equals), 0)
 
 		var tss []*state.TaskSet
-		for i, path := range paths {
-			si := infos[i]
-			c.Check(path, testutil.FileEquals, si.RealName)
+		for i, si := range infos {
+			c.Check(tmpPaths[i], testutil.FileEquals, si.RealName)
 
 			ts := state.NewTaskSet(s.NewTask("fake-install-snap", fmt.Sprintf("Doing a fake install of %q", si.RealName)))
 			tss = append(tss, ts)
@@ -855,7 +1120,7 @@ func (s *sideloadSuite) TestSideloadManySnapsAsserted(c *check.C) {
 
 		var tss []*state.TaskSet
 		for i, si := range infos {
-			c.Check(si, check.DeepEquals, &snap.SideInfo{
+			c.Check(*si, check.DeepEquals, snap.SideInfo{
 				RealName: snaps[i],
 				SnapID:   snaps[i] + "-id",
 				Revision: snap.R(41),
@@ -901,7 +1166,7 @@ func (s *sideloadSuite) TestSideloadManySnapsOneNotAsserted(c *check.C) {
 	// unasserted snap
 	twoSnap := snaptest.MakeTestSnapWithFiles(c, `name: two
 version: 1`, nil)
-	twoSnapData, err := ioutil.ReadFile(twoSnap)
+	twoSnapData, err := os.ReadFile(twoSnap)
 	c.Assert(err, check.IsNil)
 	snapData = append(snapData, twoSnapData)
 
@@ -929,7 +1194,7 @@ func (s *sideloadSuite) mockAssertions(c *check.C, st *state.State, snaps []stri
 version: 1`, snap), nil)
 		digest, size, err := asserts.SnapFileSHA3_384(thisSnap)
 		c.Assert(err, check.IsNil)
-		thisSnapData, err := ioutil.ReadFile(thisSnap)
+		thisSnapData, err := os.ReadFile(thisSnap)
 		c.Assert(err, check.IsNil)
 		snapData = append(snapData, thisSnapData)
 
@@ -980,7 +1245,7 @@ func (s *trySuite) TestTrySnap(c *check.C) {
 	snapYaml := filepath.Join(tryDir, "meta", "snap.yaml")
 	err = os.MkdirAll(filepath.Dir(snapYaml), 0755)
 	c.Assert(err, check.IsNil)
-	err = ioutil.WriteFile(snapYaml, []byte("name: foo\nversion: 1.0\n"), 0644)
+	err = os.WriteFile(snapYaml, []byte("name: foo\nversion: 1.0\n"), 0644)
 	c.Assert(err, check.IsNil)
 
 	reqForFlags := func(f snapstate.Flags) *http.Request {
@@ -1050,12 +1315,17 @@ func (s *trySuite) TestTrySnap(c *check.C) {
 			return state.NewTaskSet(t), nil
 		})()
 
-		defer daemon.MockSnapstateInstall(func(ctx context.Context, s *state.State, name string, opts *snapstate.RevisionOptions, userID int, flags snapstate.Flags) (*state.TaskSet, error) {
-			if name != "core" {
-				c.Check(flags, check.DeepEquals, t.flags, check.Commentf(t.desc))
+		defer daemon.MockSnapstateInstallWithGoal(func(ctx context.Context, st *state.State, g snapstate.InstallGoal, opts snapstate.Options) ([]*snap.Info, []*state.TaskSet, error) {
+			goal, ok := g.(*storeInstallGoalRecorder)
+			c.Assert(ok, check.Equals, true, check.Commentf("unexpected InstallGoal type %T", g))
+			c.Assert(goal.snaps, check.HasLen, 1)
+
+			if goal.snaps[0].InstanceName != "core" {
+				c.Check(opts.Flags, check.DeepEquals, t.flags, check.Commentf(t.desc))
 			}
-			t := s.NewTask("fake-install-snap", "Doing a fake install")
-			return state.NewTaskSet(t), nil
+
+			t := st.NewTask("fake-install-snap", "Doing a fake install")
+			return []*snap.Info{{}}, []*state.TaskSet{state.NewTaskSet(t)}, nil
 		})()
 
 		// try the snap (without an installed core)
@@ -1083,7 +1353,8 @@ func (s *trySuite) TestTrySnap(c *check.C) {
 		err = chg.Get("api-data", &apiData)
 		c.Assert(err, check.IsNil, check.Commentf(t.desc))
 		c.Check(apiData, check.DeepEquals, map[string]interface{}{
-			"snap-name": "foo",
+			"snap-name":  "foo",
+			"snap-names": []interface{}{"foo"},
 		}, check.Commentf(t.desc))
 
 		c.Check(soon, check.Equals, 1, check.Commentf(t.desc))
