@@ -244,7 +244,7 @@ func (s *validateGadgetTestSuite) TestValidateConsistencyWithoutModelCharacteris
 		c.Logf("tc: %v %v %v", i, tc.role, tc.label)
 		b := &bytes.Buffer{}
 
-		fmt.Fprintf(b, `
+		fmt.Fprint(b, `
 volumes:
   pc:
     bootloader: grub
@@ -252,7 +252,7 @@ volumes:
     structure:`)
 
 		if tc.role == "system-seed" {
-			fmt.Fprintf(b, `
+			fmt.Fprint(b, `
       - name: Recovery
         size: 10M
         type: 83
@@ -328,16 +328,16 @@ volumes:
 		c.Logf("tc: %v %v %v %v", i, tc.addSeed, tc.dataLabel, tc.hasModes)
 		b := &bytes.Buffer{}
 
-		fmt.Fprintf(b, bloader)
+		fmt.Fprint(b, bloader)
 		if tc.addSeed {
-			fmt.Fprintf(b, `
+			fmt.Fprint(b, `
       - name: Recovery
         size: 10M
         type: 83
         role: system-seed`)
 		}
 		if tc.addBoot {
-			fmt.Fprintf(b, `
+			fmt.Fprint(b, `
       - name: Boot
         size: 10M
         type: 83
@@ -354,7 +354,7 @@ volumes:
 		}
 
 		if tc.addSave {
-			fmt.Fprintf(b, `
+			fmt.Fprint(b, `
       - name: Save
         size: 10M
         type: 83
@@ -647,7 +647,7 @@ func (s *validateGadgetTestSuite) TestRuleValidateHybridGadgetBrokenDupRole(c *C
 	c.Check(err, ErrorMatches, `cannot have more than one partition with system-boot role`)
 }
 
-func (s *validateGadgetTestSuite) TestValidateContentMissingRawContent(c *C) {
+func (s *validateGadgetTestSuite) TestValidateContentRawContent(c *C) {
 	var gadgetYamlContent = `
 volumes:
   pc:
@@ -659,14 +659,28 @@ volumes:
         offset: 1M
         content:
           - image: foo.img
-
+            size: 1
 `
 	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlContent))
 
 	ginfo, err := gadget.ReadInfo(s.dir, nil)
 	c.Assert(err, IsNil)
 	err = gadget.ValidateContent(ginfo, s.dir, "")
-	c.Assert(err, ErrorMatches, `invalid layout of volume "pc": cannot lay out structure #0 \("foo"\): content "foo.img": stat .*/foo.img: no such file or directory`)
+	c.Assert(err, ErrorMatches, `structure #0 \("foo"\): content "foo.img": stat .*/foo.img: no such file or directory`)
+
+	// Now create the file with wrong size
+	makeSizedFile(c, filepath.Join(s.dir, "foo.img"), 100, nil)
+	ginfo, err = gadget.ReadInfo(s.dir, nil)
+	c.Assert(err, IsNil)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
+	c.Assert(err, ErrorMatches, `structure #0 \("foo"\): content "foo.img" size 100 is larger than declared 1`)
+
+	// Now with the right size
+	makeSizedFile(c, filepath.Join(s.dir, "foo.img"), 1, nil)
+	ginfo, err = gadget.ReadInfo(s.dir, nil)
+	c.Assert(err, IsNil)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
+	c.Assert(err, IsNil)
 }
 
 func (s *validateGadgetTestSuite) TestValidateContentMultiVolumeContent(c *C) {
@@ -696,7 +710,7 @@ volumes:
 	ginfo, err := gadget.ReadInfo(s.dir, nil)
 	c.Assert(err, IsNil)
 	err = gadget.ValidateContent(ginfo, s.dir, "")
-	c.Assert(err, ErrorMatches, `invalid layout of volume "second": cannot lay out structure #0 \("second-foo"\): content "second.img": stat .*/second.img: no such file or directory`)
+	c.Assert(err, ErrorMatches, `structure #0 \("second-foo"\): content "second.img": stat .*/second.img: no such file or directory`)
 }
 
 func (s *validateGadgetTestSuite) TestValidateContentFilesystemContent(c *C) {
