@@ -65,6 +65,7 @@ var templateCommon = `
 
 #include <tunables/global>
 
+###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###
 ###INCLUDE_IF_EXISTS_SNAP_TUNING###
 
 # snapd supports the concept of 'parallel installs' where snaps with the same
@@ -80,7 +81,7 @@ var templateCommon = `
 
 ###VAR###
 
-###PROFILEATTACH### (attach_disconnected,mediate_deleted) {
+###PROFILEATTACH### ###FLAGS### {
   #include <abstractions/base>
   #include <abstractions/consoles>
   #include <abstractions/openssl>
@@ -113,13 +114,7 @@ var templateCommon = `
   #include <abstractions/python>
   /etc/python3.[0-9]*/**                                r,
 
-  # explicitly deny noisy denials to read-only filesystems (see LP: #1496895
-  # for details)
-  deny /usr/lib/python3*/{,**/}__pycache__/ w,
-  deny /usr/lib/python3*/{,**/}__pycache__/**.pyc.[0-9]* w,
-  # bind mount used here (see 'parallel installs', above)
-  deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/             w,
-  deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/*.pyc.[0-9]* w,
+  ###PYCACHEDENY###
 
   # for perl apps/services
   #include <abstractions/perl>
@@ -255,6 +250,7 @@ var templateCommon = `
   /dev/{,u}random w,
   /etc/machine-id r,
   /etc/mime.types r,
+  /etc/default/keyboard r,
   @{PROC}/ r,
   @{PROC}/version r,
   @{PROC}/version_signature r,
@@ -301,6 +297,7 @@ var templateCommon = `
   /sys/fs/cgroup/memory/{,user.slice/}memory.limit_in_bytes r,
   /sys/fs/cgroup/memory/{,**/}snap.@{SNAP_INSTANCE_NAME}{,.*}/memory.limit_in_bytes r,
   /sys/fs/cgroup/memory/{,**/}snap.@{SNAP_INSTANCE_NAME}{,.*}/memory.stat r,
+  /sys/fs/cgroup/system.slice/snap.@{SNAP_INSTANCE_NAME}{,.*}/memory.max r,
   /sys/fs/cgroup/cpu,cpuacct/{,user.slice/}cpu.cfs_{period,quota}_us r,
   /sys/fs/cgroup/cpu,cpuacct/{,**/}snap.@{SNAP_INSTANCE_NAME}{,.*}/cpu.cfs_{period,quota}_us r,
   /sys/fs/cgroup/cpu,cpuacct/{,user.slice/}cpu.shares r,
@@ -377,7 +374,7 @@ var templateCommon = `
   /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/@{SNAP_REVISION}/** wl,
   /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/common/** wl,
 
-  # The ubuntu-core-launcher creates an app-specific private restricted /tmp
+  # The snap-confine program creates an app-specific private restricted /tmp
   # and will fail to launch the app if something goes wrong. As such, we can
   # simply allow full access to /tmp.
   /tmp/   r,
@@ -685,6 +682,7 @@ var defaultOtherBaseTemplateRules = `
   # - /lib/modules
   #
   # Everything but /lib/firmware and /lib/modules
+  # TODO: use GenerateAAREExclusionPatterns for this
   /{,usr/}lib/ r,
   /{,usr/}lib/[^fm]** mrklix,
   /{,usr/}lib/{f[^i],m[^o]}** mrklix,
@@ -713,6 +711,7 @@ var defaultOtherBaseTemplateRules = `
   #
   # Everything but /usr/lib and /usr/src, which are handled elsewhere.
   /usr/ r,
+  # TODO: use GenerateAAREExclusionPatterns for this
   /usr/[^ls]** mrklix,
   /usr/{l[^i],s[^r]}** mrklix,
   /usr/{li[^b],sr[^c]}** mrklix,
@@ -848,9 +847,11 @@ var coreSnippet = `
 var classicTemplate = `
 #include <tunables/global>
 
+###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###
+
 ###VAR###
 
-###PROFILEATTACH### (attach_disconnected,mediate_deleted) {
+###PROFILEATTACH### ###FLAGS### {
   # set file rules so that exec() inherits our profile unless there is
   # already a profile for it (eg, snap-confine)
   / rwkl,
@@ -886,6 +887,9 @@ var classicJailmodeSnippet = `
 
   # For snappy reexec on 4.8+ kernels
   @{INSTALL_DIR}/core/*/usr/lib/snapd/snap-exec m,
+  # Same as above but accounting for the case when the
+  # snapd snap is installed and executes the snap application.
+  @{INSTALL_DIR}/snapd/*/usr/lib/snapd/snap-exec rm,
 `
 
 var ptraceTraceDenySnippet = `
@@ -902,6 +906,16 @@ deny ptrace (trace),
 deny capability sys_ptrace,
 `
 
+var pycacheDenySnippet = `
+# explicitly deny noisy denials to read-only filesystems (see LP: #1496895
+# for details)
+deny /usr/lib/python3*/{,**/}__pycache__/ w,
+deny /usr/lib/python3*/{,**/}__pycache__/**.pyc.[0-9]* w,
+# bind mount used here (see 'parallel installs', above)
+deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/             w,
+deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/*.pyc.[0-9]* w,
+`
+
 var sysModuleCapabilityDenySnippet = `
 # The rtnetlink kernel interface can trigger the loading of kernel modules,
 # first attempting to operate on a network module (this requires the net_admin
@@ -932,6 +946,8 @@ var updateNSTemplate = `
 
 #include <tunables/global>
 
+###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###
+
 profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
   # The next four rules mirror those above. We want to be able to read
   # and map snap-update-ns into memory but it may come from a variety of places.
@@ -964,15 +980,19 @@ profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
   /sys/devices/system/cpu/online r,
 
   # Allow reading the command line (snap-update-ns uses it in pre-Go bootstrap code).
-  @{PROC}/@{pid}/cmdline r,
+  owner @{PROC}/@{pid}/cmdline r,
+
+  # Allow reading of own maps (Go runtime)
+  owner @{PROC}/@{pid}/maps r,
 
   # Allow reading file descriptor paths
-  @{PROC}/@{pid}/fd/* r,
+  owner @{PROC}/@{pid}/fd/* r,
+
   # Allow reading /proc/version. For release.go WSL detection.
   @{PROC}/version r,
 
   # Allow reading own cgroups
-  @{PROC}/@{pid}/cgroup r,
+  owner @{PROC}/@{pid}/cgroup r,
 
   # Allow reading somaxconn, required in newer distro releases
   @{PROC}/sys/net/core/somaxconn r,
@@ -987,6 +1007,12 @@ profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
   /run/snapd/lock/###SNAP_INSTANCE_NAME###.lock rwk,
   /run/snapd/lock/.lock rwk,
 
+  # While the base abstraction has rules for encryptfs encrypted home and
+  # private directories, it is missing rules for directory read on the toplevel
+  # directory of the mount (LP: #1848919)
+  owner @{HOME}/.Private/ r,
+  owner @{HOMEDIRS}/.ecryptfs/*/.Private/ r,
+
   # Allow reading stored mount namespaces,
   /run/snapd/ns/ r,
   /run/snapd/ns/###SNAP_INSTANCE_NAME###.mnt r,
@@ -1045,6 +1071,8 @@ profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
   /tmp/ r,
   /usr/ r,
   /var/ r,
+  /var/lib/ r,
+  /var/lib/snapd/ r,
   /var/snap/ r,
 
   # Allow reading timezone data.
@@ -1074,6 +1102,9 @@ profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
   /usr/lib/snapd/info r,
   /lib/apparmor/functions r,
 
+  # Allow snap-update-ns to open home directory
+  owner @{HOME}/ r,
+
 ###SNIPPETS###
 }
 `
