@@ -62,6 +62,8 @@ func (iface *boolFileInterface) StaticInfo() interfaces.StaticInfo {
 
 var boolFileGPIOValuePattern = regexp.MustCompile(
 	"^/sys/class/gpio/gpio[0-9]+/value$")
+var boolFileLedPattern = regexp.MustCompile(
+	"^/sys/devices/platform/leds/[^/]+/[^/]+/brightness$")
 var boolFileAllowedPathPatterns = []*regexp.Regexp{
 	// The brightness of standard LED class device
 	regexp.MustCompile("^/sys/class/leds/[^/]+/brightness$"),
@@ -76,6 +78,10 @@ func (iface *boolFileInterface) BeforePrepareSlot(slot *snap.SlotInfo) error {
 	if !ok || path == "" {
 		return fmt.Errorf("bool-file must contain the path attribute")
 	}
+	// XXX: this interface feeds the cleaned path into the regex and is
+	// left unchanged here for historical reasons. New interfaces (eg,
+	// like raw-volume) should instead use verifySlotPathAttribute() which
+	// performs additional verification.
 	path = filepath.Clean(path)
 	for _, pattern := range boolFileAllowedPathPatterns {
 		if pattern.MatchString(path) {
@@ -91,7 +97,6 @@ func (iface *boolFileInterface) AppArmorPermanentSlot(spec *apparmor.Specificati
 /sys/class/gpio/unexport rw,
 /sys/class/gpio/gpio[0-9]+/direction rw,
 `
-
 	if iface.isGPIO(slot) {
 		spec.AddSnippet(gpioSnippet)
 	}
@@ -108,6 +113,14 @@ func (iface *boolFileInterface) AppArmorConnectedPlug(spec *apparmor.Specificati
 		return fmt.Errorf("cannot compute plug security snippet: %v", err)
 	}
 	spec.AddSnippet(fmt.Sprintf("%s rwk,", path))
+	// Provide read/write access to trigger, delay_on and delay_off kernel files
+	if boolFileLedPattern.MatchString(path) {
+		pathDir := filepath.Dir(path)
+		spec.AddSnippet(fmt.Sprintf("%s/trigger rw,", pathDir))
+		spec.AddSnippet(fmt.Sprintf("%s/delay_on rw,", pathDir))
+		spec.AddSnippet(fmt.Sprintf("%s/delay_off rw,", pathDir))
+	}
+
 	return nil
 }
 
