@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2014-2022 Canonical Ltd
+ * Copyright (C) 2014-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,6 +20,7 @@
 package devicestate
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"runtime"
@@ -28,6 +29,7 @@ import (
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/i18n"
+	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/devicestate/internal"
 	"github.com/snapcore/snapd/overlord/snapstate"
@@ -42,7 +44,7 @@ var errNothingToDo = errors.New("nothing to do")
 
 var runtimeNumCPU = runtime.NumCPU
 
-func installSeedSnap(st *state.State, sn *seed.Snap, flags snapstate.Flags) (*state.TaskSet, *snap.Info, error) {
+func installSeedSnap(st *state.State, sn *seed.Snap, flags snapstate.Flags, prqt snapstate.PrereqTracker) (*state.TaskSet, *snap.Info, error) {
 	if sn.Required {
 		flags.Required = true
 	}
@@ -53,7 +55,21 @@ func installSeedSnap(st *state.State, sn *seed.Snap, flags snapstate.Flags) (*st
 		flags.DevMode = true
 	}
 
-	return snapstate.InstallPath(st, sn.SideInfo, sn.Path, "", sn.Channel, flags)
+	// TODO:COMPS: this will need to account for components in the future
+
+	goal := snapstate.PathInstallGoal("", sn.Path, sn.SideInfo, nil, snapstate.RevisionOptions{
+		Channel: sn.Channel,
+	})
+	info, ts, err := snapstate.InstallOne(context.Background(), st, goal, snapstate.Options{
+		Flags:         flags,
+		PrereqTracker: prqt,
+		Seed:          true,
+	})
+	if err != nil {
+		return nil, nil, err
+	}
+
+	return ts, info, nil
 }
 
 func criticalTaskEdges(ts *state.TaskSet) (beginEdge, beforeHooksEdge, hooksEdge *state.Task, err error) {
@@ -76,9 +92,11 @@ func criticalTaskEdges(ts *state.TaskSet) (beginEdge, beforeHooksEdge, hooksEdge
 
 // maybeEnforceValidationSetsTask returns a task for tracking validation-sets. This may
 // return nil if no validation-sets are present.
-func maybeEnforceValidationSetsTask(st *state.State, model *asserts.Model) (*state.Task, error) {
-	vsKey := func(accountID, name string) string {
-		return fmt.Sprintf("%s/%s", accountID, name)
+func maybeEnforceValidationSetsTask(st *state.State, model *asserts.Model, mode string) (*state.Task, error) {
+	// Only enforce validation-sets in run-mode after installing all required snaps
+	if mode != "run" {
+		logger.Debugf("Postponing enforcement of validation-sets in mode %s", mode)
+		return nil, nil
 	}
 
 	// Encode validation-sets included in the seed
@@ -95,15 +113,14 @@ func maybeEnforceValidationSetsTask(st *state.State, model *asserts.Model) (*sta
 	vsKeys := make(map[string][]string)
 	for _, a := range as {
 		vsa := a.(*asserts.ValidationSet)
-		vsKeys[vsKey(vsa.AccountID(), vsa.Name())] = a.Ref().PrimaryKey
+		vsKeys[vsa.SequenceKey()] = a.Ref().PrimaryKey
 	}
 
 	// Set up pins from the model
 	pins := make(map[string]int)
 	for _, vs := range model.ValidationSets() {
-		key := vsKey(vs.AccountID, vs.Name)
 		if vs.Sequence > 0 {
-			pins[key] = vs.Sequence
+			pins[vs.SequenceKey()] = vs.Sequence
 		}
 	}
 
@@ -255,6 +272,8 @@ func (m *DeviceManager) populateStateFromSeedImpl(tm timings.Measurer) ([]*state
 
 	infoToTs := make(map[*snap.Info]*state.TaskSet, len(essentialSeedSnaps))
 
+	prqt := snap.NewSelfContainedSetPrereqTracker()
+
 	if len(essentialSeedSnaps) != 0 {
 		// we *always* configure "core" here even if bases are used
 		// for booting. "core" is where the system config lives.
@@ -275,7 +294,7 @@ func (m *DeviceManager) populateStateFromSeedImpl(tm timings.Measurer) ([]*state
 			ApplySnapDevMode: modelIsDangerous,
 		}
 
-		ts, info, err := installSeedSnap(st, seedSnap, flags)
+		ts, info, err := installSeedSnap(st, seedSnap, flags, prqt)
 		if err != nil {
 			return nil, err
 		}
@@ -316,7 +335,7 @@ func (m *DeviceManager) populateStateFromSeedImpl(tm timings.Measurer) ([]*state
 			Classic: release.OnClassic && modelIsDangerous,
 		}
 
-		ts, info, err := installSeedSnap(st, seedSnap, flags)
+		ts, info, err := installSeedSnap(st, seedSnap, flags, prqt)
 		if err != nil {
 			return nil, err
 		}
@@ -324,12 +343,18 @@ func (m *DeviceManager) populateStateFromSeedImpl(tm timings.Measurer) ([]*state
 		infoToTs[info] = ts
 	}
 
-	// validate that all snaps have bases
-	errs := snap.ValidateBasesAndProviders(infos)
+	// validate that all snaps have bases and providers are fulfilled
+	// using the PrereqTracker
+	warns, errs := prqt.Check()
 	if errs != nil {
 		// only report the first error encountered
 		return nil, errs[0]
 	}
+	// XXX do better, use the warnings to setup checks at end of the seeding
+	// and log onlys plug not connected or explicitly disconnected there
+	for _, w := range warns {
+		logger.Noticef("seed prerequisites: %v", w)
+	}
 
 	// now add/chain the tasksets in the right order, note that we
 	// only have tasksets that we did not already seeded
@@ -345,7 +370,7 @@ func (m *DeviceManager) populateStateFromSeedImpl(tm timings.Measurer) ([]*state
 
 	// Start tracking any validation sets included in the seed after
 	// installing the included snaps.
-	if trackVss, err := maybeEnforceValidationSetsTask(st, model); err != nil {
+	if trackVss, err := maybeEnforceValidationSetsTask(st, model, mode); err != nil {
 		return nil, err
 	} else if trackVss != nil {
 		trackVss.WaitAll(ts)
@@ -411,7 +436,7 @@ func (m *DeviceManager) importAssertionsFromSeed(isCoreBoot bool) (seed.Seed, er
 		} else {
 			msg = "cannot seed a classic system with an all-snaps model"
 		}
-		return nil, fmt.Errorf(msg)
+		return nil, errors.New(msg)
 	}
 
 	// set device,model from the model assertion
