@@ -76,20 +76,26 @@ func checkSystemRequestConflict(st *state.State, systemLabel string) error {
 	return nil
 }
 
-func systemFromSeed(label string, current *currentSystem) (*System, error) {
-	s, err := seed.Open(dirs.SnapSeedDir, label)
+func systemFromSeed(label string, current *currentSystem, defaultRecoverySystem *DefaultRecoverySystem) (*System, error) {
+	_, sys, err := loadSeedAndSystem(label, current, defaultRecoverySystem)
+	return sys, err
+}
+
+func loadSeedAndSystem(label string, current *currentSystem, defaultRecoverySystem *DefaultRecoverySystem) (seed.Seed, *System, error) {
+	s, err := seedOpen(dirs.SnapSeedDir, label)
 	if err != nil {
-		return nil, fmt.Errorf("cannot open: %v", err)
+		return nil, nil, fmt.Errorf("cannot open: %v", err)
 	}
 	if err := s.LoadAssertions(nil, nil); err != nil {
-		return nil, fmt.Errorf("cannot load assertions: %v", err)
+		return nil, nil, fmt.Errorf("cannot load assertions for label %q: %v", label, err)
 	}
 	// get the model
 	model := s.Model()
 	brand, err := s.Brand()
 	if err != nil {
-		return nil, fmt.Errorf("cannot obtain brand: %v", err)
+		return nil, nil, fmt.Errorf("cannot obtain brand: %v", err)
 	}
+
 	system := &System{
 		Current: false,
 		Label:   label,
@@ -97,11 +103,12 @@ func systemFromSeed(label string, current *currentSystem) (*System, error) {
 		Brand:   brand,
 		Actions: defaultSystemActions,
 	}
+	system.DefaultRecoverySystem = defaultRecoverySystem.sameAs(system)
 	if current.sameAs(system) {
 		system.Current = true
 		system.Actions = current.actions
 	}
-	return system, nil
+	return s, system, nil
 }
 
 type currentSystem struct {
@@ -146,9 +153,6 @@ func currentSystemForMode(st *state.State, mode string) (*currentSystem, error)
 }
 
 func currentSeededSystem(st *state.State) (*seededSystem, error) {
-	st.Lock()
-	defer st.Unlock()
-
 	var whatseeded []seededSystem
 	if err := st.Get("seeded-systems", &whatseeded); err != nil {
 		return nil, err
@@ -174,7 +178,7 @@ func seededSystemFromModeenv() (*seededSystem, error) {
 		return nil, fmt.Errorf("internal error: recovery system is unset")
 	}
 
-	system, err := systemFromSeed(modeEnv.RecoverySystem, nil)
+	system, err := systemFromSeed(modeEnv.RecoverySystem, nil, nil)
 	if err != nil {
 		return nil, err
 	}
@@ -190,10 +194,11 @@ func seededSystemFromModeenv() (*seededSystem, error) {
 }
 
 // getInfoFunc is expected to return for a given snap name a snap.Info for that
-// snap and whether the snap is present is present. The second bit is relevant
-// for non-essential snaps mentioned in the model, which if present and having
-// an 'optional' presence in the model, will be added to the recovery system.
-type getSnapInfoFunc func(name string) (info *snap.Info, snapIsPresent bool, err error)
+// snap, a path on disk where the snap file can be found, and whether the snap
+// is present. The last bit is relevant for non-essential snaps mentioned in the
+// model, which if present and having an 'optional' presence in the model, will
+// be added to the recovery system.
+type getSnapInfoFunc func(name string) (info *snap.Info, path string, snapIsPresent bool, err error)
 
 // snapWriteObserveFunc is called with the recovery system directory and the
 // path to a snap file being written. The snap file may be written to a location
@@ -243,7 +248,7 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 				kind = fmt.Sprintf("non-essential but %v", nonEssentialPresence)
 			}
 		}
-		info, present, err := getInfo(name)
+		info, snapPath, present, err := getInfo(name)
 		if err != nil {
 			return fmt.Errorf("cannot obtain %v snap information: %v", kind, err)
 		}
@@ -257,7 +262,7 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 		if !present {
 			return fmt.Errorf("internal error: %v snap %q not present", kind, name)
 		}
-		if _, ok := modelSnaps[info.MountFile()]; ok {
+		if _, ok := modelSnaps[snapPath]; ok {
 			// we've already seen this snap
 			return nil
 		}
@@ -265,9 +270,9 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 		// TODO: for grade dangerous we could have a channel here which is not
 		//       the model channel, handle that here
 		optsSnaps = append(optsSnaps, &seedwriter.OptionsSnap{
-			Path: info.MountFile(),
+			Path: snapPath,
 		})
-		modelSnaps[info.MountFile()] = info
+		modelSnaps[snapPath] = info
 		return nil
 	}
 
@@ -296,19 +301,34 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 		fromDB := func(ref *asserts.Ref) (asserts.Assertion, error) {
 			return ref.Resolve(db.Find)
 		}
-		return asserts.NewFetcher(db, fromDB, save)
+
+		seqFromDB := func(ref *asserts.AtSequence) (asserts.Assertion, error) {
+			if ref.Sequence <= 0 {
+				hdrs, err := asserts.HeadersFromSequenceKey(ref.Type, ref.SequenceKey)
+				if err != nil {
+					return nil, err
+				}
+				return db.FindSequence(ref.Type, hdrs, -1, -1)
+			}
+			return ref.Resolve(db.Find)
+		}
+
+		return asserts.NewSequenceFormingFetcher(db, fromDB, seqFromDB, save)
 	}
-	f, err := w.Start(db, newFetcher)
-	if err != nil {
+
+	sf := seedwriter.MakeSeedAssertionFetcher(newFetcher)
+	if err := w.Start(db, sf); err != nil {
 		return "", err
 	}
 	// past this point the system directory is present
 
+	// TODO:COMPS: take into account local components
 	localSnaps, err := w.LocalSnaps()
 	if err != nil {
 		return recoverySystemDir, err
 	}
 
+	localARefs := make(map[*seedwriter.SeedSnap][]*asserts.Ref)
 	for _, sn := range localSnaps {
 		info, ok := modelSnaps[sn.Path]
 		if !ok {
@@ -318,9 +338,9 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 		// we have in snap.Info, but getting it this way can be
 		// expensive as we need to compute the hash, try to find a
 		// better way
-		_, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, f, db)
+		_, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, sf, db)
 		if err != nil {
-			if !asserts.IsNotFound(err) {
+			if !errors.Is(err, &asserts.NotFoundError{}) {
 				return recoverySystemDir, err
 			} else if info.SnapID != "" {
 				// snap info from state must have come
@@ -329,16 +349,21 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 				return recoverySystemDir, fmt.Errorf("internal error: no assertions for asserted snap with ID: %v", info.SnapID)
 			}
 		}
-		if err := w.SetInfo(sn, info); err != nil {
+		// TODO:COMPS: consider components
+		if err := w.SetInfo(sn, info, nil); err != nil {
 			return recoverySystemDir, err
 		}
-		sn.ARefs = aRefs
+		localARefs[sn] = aRefs
 	}
 
 	if err := w.InfoDerived(); err != nil {
 		return recoverySystemDir, err
 	}
 
+	retrieveAsserts := func(sn, _, _ *seedwriter.SeedSnap) ([]*asserts.Ref, error) {
+		return localARefs[sn], nil
+	}
+
 	for {
 		// get the list of snaps we need in this iteration
 		toDownload, err := w.SnapsToDownload()
@@ -355,7 +380,7 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 			return recoverySystemDir, fmt.Errorf("internal error: need to download snaps: %v", strings.Join(which, ", "))
 		}
 
-		complete, err := w.Downloaded()
+		complete, err := w.Downloaded(retrieveAsserts)
 		if err != nil {
 			return recoverySystemDir, err
 		}
@@ -419,7 +444,7 @@ func createSystemForModelFromValidatedSnaps(model *asserts.Model, label string,
 			bootWith.GadgetSnapOrDir = sn.Path
 		}
 	}
-	if err := boot.MakeRecoverySystemBootable(boot.InitramfsUbuntuSeedDir, recoverySystemDirInRootDir, bootWith); err != nil {
+	if err := boot.MakeRecoverySystemBootable(model, boot.InitramfsUbuntuSeedDir, recoverySystemDirInRootDir, bootWith); err != nil {
 		return recoverySystemDir, fmt.Errorf("cannot make candidate recovery system %q bootable: %v", label, err)
 	}
 	logger.Noticef("created recovery system %q", label)
