@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -24,17 +24,37 @@ import (
 	"bytes"
 	"fmt"
 	"io"
+	"strconv"
+
+	"github.com/jessevdk/go-flags"
 
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/hookstate"
-
-	"github.com/jessevdk/go-flags"
+	"github.com/snapcore/snapd/strutil"
 )
 
+type MissingContextError struct {
+	subcommand string
+}
+
+func (e *MissingContextError) Error() string {
+	return fmt.Sprintf(`cannot invoke snapctl operation commands (here %q) from outside of a snap`, e.subcommand)
+}
+
 type baseCommand struct {
 	stdout io.Writer
 	stderr io.Writer
 	c      *hookstate.Context
+	name   string
+	uid    string
+}
+
+func (c *baseCommand) setName(name string) {
+	c.name = name
+}
+
+func (c *baseCommand) setUid(uid uint32) {
+	c.uid = strconv.FormatUint(uint64(uid), 10)
 }
 
 func (c *baseCommand) setStdout(w io.Writer) {
@@ -42,8 +62,12 @@ func (c *baseCommand) setStdout(w io.Writer) {
 }
 
 func (c *baseCommand) printf(format string, a ...interface{}) {
+	c.print(fmt.Sprintf(format, a...))
+}
+
+func (c *baseCommand) print(msg string) {
 	if c.stdout != nil {
-		fmt.Fprintf(c.stdout, format, a...)
+		fmt.Fprint(c.stdout, msg)
 	}
 }
 
@@ -51,9 +75,9 @@ func (c *baseCommand) setStderr(w io.Writer) {
 	c.stderr = w
 }
 
-func (c *baseCommand) errorf(format string, a ...interface{}) {
+func (c *baseCommand) error(msg string) {
 	if c.stderr != nil {
-		fmt.Fprintf(c.stderr, format, a...)
+		fmt.Fprint(c.stderr, msg)
 	}
 }
 
@@ -65,7 +89,17 @@ func (c *baseCommand) context() *hookstate.Context {
 	return c.c
 }
 
+func (c *baseCommand) ensureContext() (context *hookstate.Context, err error) {
+	if c.c == nil {
+		err = &MissingContextError{c.name}
+	}
+	return c.c, err
+}
+
 type command interface {
+	setName(name string)
+	setUid(uid uint32)
+
 	setStdout(w io.Writer)
 	setStderr(w io.Writer)
 
@@ -112,37 +146,34 @@ func (f ForbiddenCommandError) Error() string {
 	return f.Message
 }
 
-// ForbiddenCommand contains information about an attempt to use a command in a context where it is not allowed.
-type ForbiddenCommand struct {
-	Uid  uint32
-	Name string
-}
-
-func (f *ForbiddenCommand) Execute(args []string) error {
-	return &ForbiddenCommandError{Message: fmt.Sprintf("cannot use %q with uid %d, try with sudo", f.Name, f.Uid)}
-}
+// nonRootAllowed lists the commands that can be performed even when snapctl
+// is invoked not by root.
+var nonRootAllowed = []string{"get", "services", "set-health", "is-connected", "system-mode", "model"}
 
 // Run runs the requested command.
 func Run(context *hookstate.Context, args []string, uid uint32) (stdout, stderr []byte, err error) {
+	if len(args) == 0 {
+		return nil, nil, fmt.Errorf("internal error: snapctl cannot run without args")
+	}
+
+	if !isAllowedToRun(uid, args) {
+		return nil, nil, &ForbiddenCommandError{Message: fmt.Sprintf("cannot use %q with uid %d, try with sudo", args[0], uid)}
+	}
+
 	parser := flags.NewNamedParser("snapctl", flags.PassDoubleDash|flags.HelpFlag)
 
 	// Create stdout/stderr buffers, and make sure commands use them.
 	var stdoutBuffer bytes.Buffer
 	var stderrBuffer bytes.Buffer
 	for name, cmdInfo := range commands {
-		var data interface{}
-		// commands listed here will be allowed for regular users
-		// note: commands still need valid context and snaps can only access own config.
-		if uid == 0 || name == "get" || name == "services" || name == "set-health" || name == "is-connected" {
-			cmd := cmdInfo.generator()
-			cmd.setStdout(&stdoutBuffer)
-			cmd.setStderr(&stderrBuffer)
-			cmd.setContext(context)
-			data = cmd
-		} else {
-			data = &ForbiddenCommand{Uid: uid, Name: name}
-		}
-		theCmd, err := parser.AddCommand(name, cmdInfo.shortHelp, cmdInfo.longHelp, data)
+		cmd := cmdInfo.generator()
+		cmd.setName(name)
+		cmd.setUid(uid)
+		cmd.setStdout(&stdoutBuffer)
+		cmd.setStderr(&stderrBuffer)
+		cmd.setContext(context)
+
+		theCmd, err := parser.AddCommand(name, cmdInfo.shortHelp, cmdInfo.longHelp, cmd)
 		theCmd.Hidden = cmdInfo.hidden
 		if err != nil {
 			logger.Panicf("cannot add command %q: %s", name, err)
@@ -152,3 +183,37 @@ func Run(context *hookstate.Context, args []string, uid uint32) (stdout, stderr
 	_, err = parser.ParseArgs(args)
 	return stdoutBuffer.Bytes(), stderrBuffer.Bytes(), err
 }
+
+// isAllowedToRun returns true if the user with the given UID can run the given snapctl command vector.
+//
+// Commands still need valid context and snaps can only access own config.
+func isAllowedToRun(uid uint32, args []string) bool {
+	// Root can run all snapctl commands.
+	if uid == 0 {
+		return true
+	}
+
+	for idx, arg := range args {
+		// A number of sub-commands are allowed to be executed by non-root users.
+		if idx == 0 && strutil.ListContains(nonRootAllowed, arg) {
+			return true
+		}
+
+		// Invoking help is always allowed.
+		if arg == "-h" || arg == "--help" {
+			return true
+		}
+
+		// Note that we are not interrupting parsing after the first non-option
+		// argument (POSIX style), because we want to cater to the use case of
+		// the user appending --help or -h at the end of the command and still
+		// getting something useful. The only exception is the condition below.
+
+		// The explicit termination argument terminates parsing.
+		if arg == "--" {
+			break
+		}
+	}
+
+	return false
+}
