@@ -34,14 +34,13 @@ import (
 	"gopkg.in/tomb.v2"
 
 	"github.com/snapcore/snapd/dirs"
-	"github.com/snapcore/snapd/errtracker"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
-	"github.com/snapcore/snapd/overlord/configstate/settings"
 	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 )
 
 type hijackFunc func(ctx *Context) error
@@ -82,14 +81,28 @@ type HandlerGenerator func(*Context) Handler
 
 // HookSetup is a reference to a hook within a specific snap.
 type HookSetup struct {
-	Snap        string        `json:"snap"`
-	Revision    snap.Revision `json:"revision"`
-	Hook        string        `json:"hook"`
-	Timeout     time.Duration `json:"timeout,omitempty"`
-	Optional    bool          `json:"optional,omitempty"`     // do not error if script is missing
-	Always      bool          `json:"always,omitempty"`       // run handler even if script is missing
-	IgnoreError bool          `json:"ignore-error,omitempty"` // do not run handler's Error() on error
-	TrackError  bool          `json:"track-error,omitempty"`  // report hook error to oopsie
+	Snap     string        `json:"snap"`
+	Revision snap.Revision `json:"revision"`
+	Hook     string        `json:"hook"`
+	Timeout  time.Duration `json:"timeout,omitempty"`
+
+	// Optional is true if we should not error if the script is missing.
+	Optional bool `json:"optional,omitempty"`
+
+	// Always is true if we should run the handler even if the script is
+	// missing.
+	Always bool `json:"always,omitempty"`
+
+	// IgnoreError is true if we should not run the handler's Error() on error.
+	IgnoreError bool `json:"ignore-error,omitempty"`
+
+	// Component is the component name that the hook is associated with. If the
+	// hook is not associated with a component, the string will be empty.
+	Component string `json:"component,omitempty"`
+
+	// ComponentRevision is the revision of the component that the hook is
+	// associated with. Only valid if Component is not empty.
+	ComponentRevision snap.Revision `json:"component-revision"`
 }
 
 // Manager returns a new HookManager.
@@ -331,8 +344,12 @@ func (m *HookManager) runHookGuardForRestarting(context *Context) error {
 }
 
 func (m *HookManager) runHook(context *Context, snapst *snapstate.SnapState, hooksup *HookSetup, tomb *tomb.Tomb) error {
-	mustHijack := m.hijacked(hooksup.Hook, hooksup.Snap) != nil
+	// for now, we will only support hijacking snap hooks, not component hooks.
+	// if we ever add components to the snapd snap, we might need to handle
+	// hijacking component hooks as well.
+	mustHijack := context.IsSnapHook() && m.hijacked(hooksup.Hook, hooksup.Snap) != nil
 	hookExists := false
+
 	if !mustHijack {
 		// not hijacked, snap must be installed
 		if !snapst.IsInstalled() {
@@ -344,9 +361,24 @@ func (m *HookManager) runHook(context *Context, snapst *snapstate.SnapState, hoo
 			return fmt.Errorf("cannot read %q snap details: %v", hooksup.Snap, err)
 		}
 
-		hookExists = info.Hooks[hooksup.Hook] != nil
-		if !hookExists && !hooksup.Optional {
-			return fmt.Errorf("snap %q has no %q hook", hooksup.Snap, hooksup.Hook)
+		if context.IsSnapHook() {
+			hookExists = info.Hooks[hooksup.Hook] != nil
+			if !hookExists && !hooksup.Optional {
+				return fmt.Errorf("snap %q has no %q hook", hooksup.Snap, hooksup.Hook)
+			}
+		} else {
+			comp, err := snapst.CurrentComponentInfo(naming.ComponentRef{
+				SnapName:      info.SnapName(),
+				ComponentName: hooksup.Component,
+			})
+			if err != nil {
+				return fmt.Errorf(`cannot read "%s+%s" component details: %v`, info.SnapName(), hooksup.Component, err)
+			}
+
+			hookExists = comp.Hooks[hooksup.Hook] != nil
+			if !hookExists && !hooksup.Optional {
+				return fmt.Errorf(`component "%s+%s" has no %q hook`, info.SnapName(), hooksup.Component, hooksup.Hook)
+			}
 		}
 	}
 
@@ -404,9 +436,7 @@ func (m *HookManager) runHook(context *Context, snapst *snapstate.SnapState, hoo
 		output, err = runHook(context, tomb)
 	}
 	if err != nil {
-		if hooksup.TrackError {
-			trackHookError(context, output, err)
-		}
+		// TODO: telemetry about errors here
 		err = osutil.OutputErr(output, err)
 		if hooksup.IgnoreError {
 			context.Lock()
@@ -439,7 +469,7 @@ func (m *HookManager) runHook(context *Context, snapst *snapstate.SnapState, hoo
 }
 
 func runHookImpl(c *Context, tomb *tomb.Tomb) ([]byte, error) {
-	return runHookAndWait(c.InstanceName(), c.SnapRevision(), c.HookName(), c.ID(), c.Timeout(), tomb)
+	return runHookAndWait(c.HookSource(), c.SnapRevision(), c.HookName(), c.ID(), c.Timeout(), tomb)
 }
 
 var runHook = runHookImpl
@@ -478,8 +508,8 @@ func snapCmd() string {
 
 var defaultHookTimeout = 10 * time.Minute
 
-func runHookAndWait(snapName string, revision snap.Revision, hookName, hookContext string, timeout time.Duration, tomb *tomb.Tomb) ([]byte, error) {
-	argv := []string{snapCmd(), "run", "--hook", hookName, "-r", revision.String(), snapName}
+func runHookAndWait(hookSource string, revision snap.Revision, hookName, hookContext string, timeout time.Duration, tomb *tomb.Tomb) ([]byte, error) {
+	argv := []string{snapCmd(), "run", "--hook", hookName, "-r", revision.String(), hookSource}
 	if timeout == 0 {
 		timeout = defaultHookTimeout
 	}
@@ -496,28 +526,3 @@ func runHookAndWait(snapName string, revision snap.Revision, hookName, hookConte
 
 	return osutil.RunAndWait(argv, env, timeout, tomb)
 }
-
-var errtrackerReport = errtracker.Report
-
-func trackHookError(context *Context, output []byte, err error) {
-	errmsg := fmt.Sprintf("hook %s in snap %q failed: %v", context.HookName(), context.InstanceName(), osutil.OutputErr(output, err))
-	dupSig := fmt.Sprintf("hook:%s:%s:%s\n%s", context.InstanceName(), context.HookName(), err, output)
-	extra := map[string]string{
-		"HookName": context.HookName(),
-	}
-	if context.setup.IgnoreError {
-		extra["IgnoreError"] = "1"
-	}
-
-	context.state.Lock()
-	problemReportsDisabled := settings.ProblemReportsDisabled(context.state)
-	context.state.Unlock()
-	if !problemReportsDisabled {
-		oopsid, err := errtrackerReport(context.InstanceName(), errmsg, dupSig, extra)
-		if err == nil {
-			logger.Noticef("Reported hook failure from %q for snap %q as %s", context.HookName(), context.InstanceName(), oopsid)
-		} else {
-			logger.Debugf("Cannot report hook failure: %s", err)
-		}
-	}
-}
