@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2017 Canonical Ltd
+ * Copyright (C) 2017-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -34,6 +34,9 @@ import (
 
 func init() {
 	snapstate.SetupInstallHook = SetupInstallHook
+	snapstate.SetupInstallComponentHook = SetupInstallComponentHook
+	snapstate.SetupPreRefreshComponentHook = SetupPreRefreshComponentHook
+	snapstate.SetupPostRefreshComponentHook = SetupPostRefreshComponentHook
 	snapstate.SetupPreRefreshHook = SetupPreRefreshHook
 	snapstate.SetupPostRefreshHook = SetupPostRefreshHook
 	snapstate.SetupRemoveHook = SetupRemoveHook
@@ -53,6 +56,48 @@ func SetupInstallHook(st *state.State, snapName string) *state.Task {
 	return task
 }
 
+func SetupInstallComponentHook(st *state.State, snap, component string) *state.Task {
+	hooksup := &HookSetup{
+		Snap:      snap,
+		Component: component,
+		Hook:      "install",
+		Optional:  true,
+	}
+
+	summary := fmt.Sprintf(i18n.G(`Run install hook of "%s+%s" component if present`), hooksup.Snap, hooksup.Component)
+	task := HookTask(st, summary, hooksup, nil)
+
+	return task
+}
+
+func SetupPreRefreshComponentHook(st *state.State, snap, component string) *state.Task {
+	hooksup := &HookSetup{
+		Snap:      snap,
+		Component: component,
+		Hook:      "pre-refresh",
+		Optional:  true,
+	}
+
+	summary := fmt.Sprintf(i18n.G(`Run pre-refresh hook of "%s+%s" component if present`), hooksup.Snap, hooksup.Component)
+	task := HookTask(st, summary, hooksup, nil)
+
+	return task
+}
+
+func SetupPostRefreshComponentHook(st *state.State, snap, component string) *state.Task {
+	hooksup := &HookSetup{
+		Snap:      snap,
+		Component: component,
+		Hook:      "post-refresh",
+		Optional:  true,
+	}
+
+	summary := fmt.Sprintf(i18n.G(`Run post-refresh hook of "%s+%s" component if present`), hooksup.Snap, hooksup.Component)
+	task := HookTask(st, summary, hooksup, nil)
+
+	return task
+}
+
 func SetupPostRefreshHook(st *state.State, snapName string) *state.Task {
 	hooksup := &HookSetup{
 		Snap:     snapName,
@@ -99,6 +144,11 @@ func (h *gateAutoRefreshHookHandler) Before() error {
 	h.refreshAppAwareness = true
 
 	snapName := h.context.InstanceName()
+	snapInfo, err := snapstate.CurrentInfo(st, snapName)
+	if err != nil {
+		return err
+	}
+	snapRev := snapInfo.SnapRevision()
 
 	// obtain snap lock before manipulating runinhibit lock.
 	lock, err := snaplock.OpenLock(snapName)
@@ -110,7 +160,8 @@ func (h *gateAutoRefreshHookHandler) Before() error {
 	}
 	defer lock.Unlock()
 
-	if err := runinhibit.LockWithHint(snapName, runinhibit.HintInhibitedGateRefresh); err != nil {
+	inhibitInfo := runinhibit.InhibitInfo{Previous: snapRev}
+	if err := runinhibit.LockWithHint(snapName, runinhibit.HintInhibitedGateRefresh, inhibitInfo); err != nil {
 		return err
 	}
 
@@ -123,7 +174,7 @@ func (h *gateAutoRefreshHookHandler) Done() (err error) {
 	ctx.Lock()
 	defer ctx.Unlock()
 
-	snapName := h.context.InstanceName()
+	snapName := ctx.InstanceName()
 
 	var action snapstate.GateAutoRefreshAction
 	a := ctx.Cached("action")
@@ -151,7 +202,7 @@ func (h *gateAutoRefreshHookHandler) Done() (err error) {
 				return fmt.Errorf("cannot unlock inhibit lock for snap %s: %v", snapName, err)
 			}
 		}
-		return snapstate.ProceedWithRefresh(st, snapName)
+		return snapstate.ProceedWithRefresh(st, snapName, nil)
 	} else {
 		var ok bool
 		action, ok = a.(snapstate.GateAutoRefreshAction)
@@ -172,13 +223,18 @@ func (h *gateAutoRefreshHookHandler) Done() (err error) {
 	case snapstate.GateAutoRefreshProceed:
 		// for action=proceed the ctlcmd doesn't call ProceedWithRefresh
 		// immediately, do it here.
-		if err := snapstate.ProceedWithRefresh(st, snapName); err != nil {
+		if err := snapstate.ProceedWithRefresh(st, snapName, nil); err != nil {
 			return err
 		}
 		if h.refreshAppAwareness {
-			// we have HintInhibitedGateRefresh lock already when running the hook,
-			// change it to HintInhibitedForRefresh.
-			if err := runinhibit.LockWithHint(snapName, runinhibit.HintInhibitedForRefresh); err != nil {
+			// we have HintInhibitedGateRefresh lock already when running the hook, change
+			// it to HintInhibitedForRefresh.
+			// Also let's reuse inhibit info that was saved in Before().
+			_, inhibitInfo, err := runinhibit.IsLocked(snapName)
+			if err != nil {
+				return err
+			}
+			if err := runinhibit.LockWithHint(snapName, runinhibit.HintInhibitedForRefresh, inhibitInfo); err != nil {
 				return fmt.Errorf("cannot set inhibit lock for snap %s: %v", snapName, err)
 			}
 		}
@@ -241,7 +297,7 @@ func (h *gateAutoRefreshHookHandler) Error(hookErr error) (ignoreHookErr bool, e
 
 	// no duration specified, use maximum allowed for this gating snap.
 	var holdDuration time.Duration
-	if _, err := snapstate.HoldRefresh(st, snapName, holdDuration, affecting...); err != nil {
+	if _, err := snapstate.HoldRefresh(st, snapstate.HoldAutoRefresh, snapName, holdDuration, affecting...); err != nil {
 		// log the original hook error as we either ignore it or error out from
 		// this handler, in both cases hookErr won't be logged by hook manager.
 		h.context.Errorf("error: %v (while handling previous hook error: %v)", err, hookErr)
