@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2014-2020 Canonical Ltd
+ * Copyright (C) 2014-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -30,6 +30,7 @@ package seed
 */
 
 import (
+	"errors"
 	"fmt"
 	"path/filepath"
 
@@ -50,6 +51,10 @@ type seed16 struct {
 
 	model *asserts.Model
 
+	yamlSnaps []*internal.Snap16
+
+	essCache map[string]*Snap
+
 	snaps             []*Snap
 	essentialSnapsNum int
 
@@ -60,7 +65,7 @@ func (s *seed16) LoadAssertions(db asserts.RODatabase, commitTo func(*asserts.Ba
 	if db == nil {
 		// a db was not provided, create an internal temporary one
 		var err error
-		db, commitTo, err = newMemAssertionsDB()
+		db, commitTo, err = newMemAssertionsDB(nil)
 		if err != nil {
 			return err
 		}
@@ -116,45 +121,92 @@ func (s *seed16) Brand() (*asserts.Account, error) {
 	return findBrand(s, s.db)
 }
 
-func (s *seed16) addSnap(sn *internal.Snap16, pinnedTrack string, tm timings.Measurer) (*Snap, error) {
+func (s *seed16) SetParallelism(int) {
+	// ignored
+}
+
+func (s *seed16) addSnap(sn *internal.Snap16, pinnedTrack string, handler ContainerHandler, cache map[string]*Snap, tm timings.Measurer) (*Snap, error) {
 	path := filepath.Join(s.seedDir, "snaps", sn.File)
-	snapChannel := sn.Channel
-	if pinnedTrack != "" {
-		var err error
-		snapChannel, err = channel.ResolvePinned(pinnedTrack, snapChannel)
-		if err != nil {
-			// fallback to using the pinned track directly
-			snapChannel = pinnedTrack
-		}
-	}
-	seedSnap := &Snap{
-		Path:    path,
-		Channel: snapChannel,
-		Classic: sn.Classic,
-		DevMode: sn.DevMode,
-	}
 
-	var sideInfo snap.SideInfo
-	if sn.Unasserted {
-		sideInfo.RealName = sn.Name
-	} else {
-		var si *snap.SideInfo
-		var err error
-		timings.Run(tm, "derive-side-info", fmt.Sprintf("hash and derive side info for snap %q", sn.Name), func(nested timings.Measurer) {
-			si, err = snapasserts.DeriveSideInfo(path, s.db)
-		})
-		if asserts.IsNotFound(err) {
-			return nil, fmt.Errorf("cannot find signatures with metadata for snap %q (%q)", sn.Name, path)
-		}
-		if err != nil {
-			return nil, err
-		}
-		sideInfo = *si
-		sideInfo.Private = sn.Private
-		sideInfo.Contact = sn.Contact
-	}
+	_, defaultHandler := handler.(defaultSnapHandler)
 
-	seedSnap.SideInfo = &sideInfo
+	seedSnap := cache[path]
+	// not cached, or ignore the cache if a non-default handler
+	// was passed, otherwise it would not be called which could be
+	// unexpected
+	if seedSnap == nil || !defaultHandler {
+		snapChannel := sn.Channel
+		if pinnedTrack != "" {
+			var err error
+			snapChannel, err = channel.ResolvePinned(pinnedTrack, snapChannel)
+			if err != nil {
+				// fallback to using the pinned track directly
+				snapChannel = pinnedTrack
+			}
+		}
+		seedSnap = &Snap{
+			Channel: snapChannel,
+			Classic: sn.Classic,
+			DevMode: sn.DevMode,
+		}
+
+		var sideInfo snap.SideInfo
+		var newPath string
+		if sn.Unasserted {
+			var err error
+			pinfo := snap.MinimalSnapContainerPlaceInfo(sn.Name, snap.Revision{N: -1})
+			newPath, err = handler.HandleUnassertedContainer(pinfo, path, tm)
+			if err != nil {
+				return nil, err
+			}
+			sideInfo.RealName = sn.Name
+		} else {
+			var si *snap.SideInfo
+			var err error
+
+			deriveRev := func(snapSHA3_384 string, snapSize uint64) (snap.Revision, error) {
+				if si == nil {
+					var err error
+					si, err = snapasserts.DeriveSideInfoFromDigestAndSize(path, snapSHA3_384, snapSize, s.model, s.db)
+					if err != nil {
+						return snap.Revision{}, err
+					}
+				}
+				return si.Revision, nil
+			}
+			timings.Run(tm, "derive-side-info", fmt.Sprintf("hash and derive side info for snap %q", sn.Name), func(nested timings.Measurer) {
+				var snapSHA3_384 string
+				var snapSize uint64
+				// NOTE: this revision is not really correct
+				cpi := snap.MinimalSnapContainerPlaceInfo(sn.Name, snap.R(0))
+				newPath, snapSHA3_384, snapSize, err = handler.HandleAndDigestAssertedContainer(cpi, path, tm)
+				if err != nil {
+					return
+				}
+				// sets si too
+				_, err = deriveRev(snapSHA3_384, snapSize)
+			})
+			if errors.Is(err, &asserts.NotFoundError{}) {
+				return nil, fmt.Errorf("cannot find signatures with metadata for snap %q (%q)", sn.Name, path)
+			}
+			if err != nil {
+				return nil, err
+			}
+			sideInfo = *si
+			sideInfo.Private = sn.Private
+			sideInfo.LegacyEditedContact = sn.Contact
+		}
+		origPath := path
+		if newPath != "" {
+			path = newPath
+		}
+		seedSnap.Path = path
+
+		seedSnap.SideInfo = &sideInfo
+		if cache != nil {
+			cache[origPath] = seedSnap
+		}
+	}
 
 	s.snaps = append(s.snaps, seedSnap)
 
@@ -169,8 +221,10 @@ func (e *essentialSnapMissingError) Error() string {
 	return fmt.Sprintf("essential snap %q required by the model is missing in the seed", e.SnapName)
 }
 
-func (s *seed16) LoadMeta(tm timings.Measurer) error {
-	model := s.Model()
+func (s *seed16) loadYaml() error {
+	if s.yamlSnaps != nil {
+		return nil
+	}
 
 	seedYamlFile := filepath.Join(s.seedDir, "seed.yaml")
 	if !osutil.FileExists(seedYamlFile) {
@@ -181,14 +235,30 @@ func (s *seed16) LoadMeta(tm timings.Measurer) error {
 	if err != nil {
 		return err
 	}
-	yamlSnaps := seedYaml.Snaps
+	s.yamlSnaps = seedYaml.Snaps
 
-	required := naming.NewSnapSet(model.RequiredWithEssentialSnaps())
-	seeding := make(map[string]*internal.Snap16, len(yamlSnaps))
-	for _, sn := range yamlSnaps {
+	return nil
+}
+
+func (s *seed16) resetSnaps() {
+	// setup essential snaps cache
+	if s.essCache == nil {
+		// 4 = snapd+base+kernel+gadget
+		s.essCache = make(map[string]*Snap, 4)
+	}
+
+	s.snaps = nil
+	s.essentialSnapsNum = 0
+}
+
+func (s *seed16) loadEssentialMeta(essentialTypes []snap.Type, required *naming.SnapSet, handler ContainerHandler, added map[string]bool, tm timings.Measurer) error {
+	model := s.Model()
+
+	seeding := make(map[string]*internal.Snap16, len(s.yamlSnaps))
+	for _, sn := range s.yamlSnaps {
 		seeding[sn.Name] = sn
 	}
-	added := make(map[string]bool, 3)
+
 	classic := model.Classic()
 	_, usesSnapdSnap := seeding["snapd"]
 	usesSnapdSnap = usesSnapdSnap || required.Contains(naming.Snap("snapd"))
@@ -212,12 +282,27 @@ func (s *seed16) LoadMeta(tm timings.Measurer) error {
 		if added[snapName] {
 			return nil, nil
 		}
+
+		// filter if required
+		if essentialTypes != nil {
+			skip := true
+			for _, t := range essentialTypes {
+				if t == essType {
+					skip = false
+					break
+				}
+			}
+			if skip {
+				return nil, nil
+			}
+		}
+
 		yamlSnap := seeding[snapName]
 		if yamlSnap == nil {
 			return nil, &essentialSnapMissingError{SnapName: snapName}
 		}
 
-		seedSnap, err := s.addSnap(yamlSnap, pinnedTrack, tm)
+		seedSnap, err := s.addSnap(yamlSnap, pinnedTrack, handler, s.essCache, tm)
 		if err != nil {
 			return nil, err
 		}
@@ -235,7 +320,7 @@ func (s *seed16) LoadMeta(tm timings.Measurer) error {
 	}
 
 	// if there are snaps to seed, core/base needs to be seeded too
-	if len(yamlSnaps) != 0 {
+	if len(s.yamlSnaps) != 0 {
 		// ensure "snapd" snap is installed first
 		if model.Base() != "" || classicWithSnapd {
 			if _, err := addEssential("snapd", "", snap.TypeSnapd); err != nil {
@@ -260,35 +345,91 @@ func (s *seed16) LoadMeta(tm timings.Measurer) error {
 		if err != nil {
 			return err
 		}
+		// not skipped
+		if gadget != nil {
 
-		// always make sure the base of gadget is installed first
-		info, err := readInfo(gadget.Path, gadget.SideInfo)
-		if err != nil {
-			return err
-		}
-		gadgetBase := info.Base
-		if gadgetBase == "" {
-			gadgetBase = "core"
-		}
-		// Sanity check
-		// TODO: do we want to relax this? the new logic would allow
-		// but it might just be confusing for now
-		if baseSnap != "" && gadgetBase != baseSnap {
-			return fmt.Errorf("cannot use gadget snap because its base %q is different from model base %q", gadgetBase, model.Base())
-		}
-		if _, err = addEssential(gadgetBase, "", snap.TypeBase); err != nil {
-			return err
+			// always make sure the base of gadget is installed first
+			info, err := readInfo(gadget.Path, gadget.SideInfo)
+			if err != nil {
+				return err
+			}
+			gadgetBase := info.Base
+			if gadgetBase == "" {
+				gadgetBase = "core"
+			}
+			// Validity check
+			// TODO: do we want to relax this? the new logic would allow
+			// but it might just be confusing for now
+			if baseSnap != "" && gadgetBase != baseSnap {
+				return fmt.Errorf("cannot use gadget snap because its base %q is different from model base %q", gadgetBase, model.Base())
+			}
+			if _, err = addEssential(gadgetBase, "", snap.TypeBase); err != nil {
+				return err
+			}
 		}
 	}
 
 	s.essentialSnapsNum = len(s.snaps)
 
+	return nil
+}
+
+func (s *seed16) LoadEssentialMeta(essentialTypes []snap.Type, tm timings.Measurer) error {
+	return s.LoadEssentialMetaWithSnapHandler(essentialTypes, nil, tm)
+}
+
+func (s *seed16) LoadEssentialMetaWithSnapHandler(essentialTypes []snap.Type, handler ContainerHandler, tm timings.Measurer) error {
+	model := s.Model()
+
+	if err := s.loadYaml(); err != nil {
+		return err
+	}
+
+	required := naming.NewSnapSet(model.RequiredWithEssentialSnaps())
+	added := make(map[string]bool, 3)
+
+	s.resetSnaps()
+
+	if handler == nil {
+		handler = defaultSnapHandler{}
+	}
+
+	if len(essentialTypes) == 0 {
+		essentialTypes = nil
+	}
+	return s.loadEssentialMeta(essentialTypes, required, handler, added, tm)
+}
+
+func (s *seed16) LoadMeta(mode string, handler ContainerHandler, tm timings.Measurer) error {
+	if mode != AllModes && mode != "run" {
+		return fmt.Errorf("internal error: Core 16/18 have only run mode, got: %s", mode)
+	}
+
+	model := s.Model()
+
+	if err := s.loadYaml(); err != nil {
+		return err
+	}
+
+	required := naming.NewSnapSet(model.RequiredWithEssentialSnaps())
+	added := make(map[string]bool, 3)
+
+	s.resetSnaps()
+
+	if handler == nil {
+		handler = defaultSnapHandler{}
+	}
+
+	if err := s.loadEssentialMeta(nil, required, handler, added, tm); err != nil {
+		return err
+	}
+
 	// the rest of the snaps
-	for _, sn := range yamlSnaps {
+	for _, sn := range s.yamlSnaps {
 		if added[sn.Name] {
 			continue
 		}
-		seedSnap, err := s.addSnap(sn, "", tm)
+		seedSnap, err := s.addSnap(sn, "", handler, nil, tm)
 		if err != nil {
 			return err
 		}
