@@ -39,6 +39,7 @@ import (
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/gadget/quantity"
+	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/osutil/squashfs"
 	"github.com/snapcore/snapd/sandbox/selinux"
@@ -118,12 +119,11 @@ func MockNewSystemd(f func(be Backend, rootDir string, mode InstanceMode, rep Re
 
 // systemctlCmd calls systemctl with the given args, returning its standard output (and wrapped error)
 var systemctlCmd = func(args ...string) ([]byte, error) {
-	// TODO: including stderr here breaks many things when systemd is in debug
-	// output mode, see LP #1885597
-	bs, err := exec.Command("systemctl", args...).CombinedOutput()
+	bs, stderr, err := osutil.RunSplitOutput("systemctl", args...)
 	if err != nil {
 		exitCode, runErr := osutil.ExitCode(err)
-		return nil, &Error{cmd: args, exitCode: exitCode, runErr: runErr, msg: bs}
+		return nil, &Error{cmd: args, exitCode: exitCode, runErr: runErr,
+			msg: osutil.CombineStdOutErr(bs, stderr)}
 	}
 
 	return bs, nil
@@ -230,6 +230,12 @@ func getVersion() (int, error) {
 		}
 	}
 
+	// ignore the pre-release suffixes, since we otherwise we can't convert to an
+	// int and we only compare versions coarsely to check systemd is not too old
+	if i := strings.IndexRune(verstr, '~'); i != -1 {
+		verstr = verstr[:i]
+	}
+
 	ver, err := strconv.Atoi(verstr)
 	if err != nil {
 		return 0, fmt.Errorf("cannot convert systemd version to number: %s", verstr)
@@ -302,16 +308,31 @@ func MockJournalctl(f func(svcs []string, n int, follow, namespaces bool) (io.Re
 	}
 }
 
+// MountUnitType is an enum for the supported mount unit types.
+type MountUnitType int
+
+const (
+	// RegularMountUnit is a mount with the usual dependencies
+	RegularMountUnit MountUnitType = iota
+	// BeforeDriversLoadMountUnit mounts things before kernel modules are
+	// loaded either by udevd or by systemd-modules-load.service.
+	BeforeDriversLoadMountUnit
+)
+
 type MountUnitOptions struct {
+	// MountUnitType is the type of mount unit we want
+	MountUnitType MountUnitType
 	// Whether the unit is transient or persistent across reboots
-	Lifetime UnitLifetime
-	SnapName string
-	Revision string
-	What     string
-	Where    string
-	Fstype   string
-	Options  []string
-	Origin   string
+	Lifetime    UnitLifetime
+	Description string
+	What        string
+	Where       string
+	Fstype      string
+	Options     []string
+	Origin      string
+	// PreventRestartIfModified is set if we do not want to restart the
+	// mount unit if modified
+	PreventRestartIfModified bool
 }
 
 // Backend identifies the implementation backend in use by a Systemd instance.
@@ -326,6 +347,26 @@ const (
 	EmulationModeBackend
 )
 
+type mountUpdateStatus int
+
+const (
+	mountUnchanged mountUpdateStatus = iota
+	mountUpdated
+	mountCreated
+)
+
+// EnsureMountUnitFlags contains flags that modify behavior of EnsureMountUnitFile
+// TODO should we call directly EnsureMountUnitFileWithOptions and
+// remove this type instead?
+type EnsureMountUnitFlags struct {
+	// PreventRestartIfModified is set if we do not want to restart the
+	// mount unit if even though it was modified
+	PreventRestartIfModified bool
+	// StartBeforeDriversLoad is set if the unit is needed before
+	// udevd starts to run rules
+	StartBeforeDriversLoad bool
+}
+
 // Systemd exposes a minimal interface to manage systemd via the systemctl command.
 type Systemd interface {
 	// Backend returns the underlying implementation backend.
@@ -351,9 +392,10 @@ type Systemd interface {
 	// Restart the service, waiting for it to stop before starting it again.
 	Restart(services []string) error
 	// Reload or restart the service via 'systemctl reload-or-restart'
-	ReloadOrRestart(service string) error
-	// RestartAll restarts the given service using systemctl restart --all
-	RestartAll(service string) error
+	ReloadOrRestart(services []string) error
+	// RestartNoWaitForStop restarts the given services using systemctl restart,
+	// with no snapd specific logic to wait for the services to stop.
+	RestartNoWaitForStop(services []string) error
 	// Status fetches the status of given units. Statuses are
 	// returned in the same order as unit names passed in
 	// argument.
@@ -379,15 +421,17 @@ type Systemd interface {
 	// If namespaces is set to true, the log reader will include journal namespace
 	// logs, and is required to get logs for services which are in journal namespaces.
 	LogReader(services []string, n int, follow, namespaces bool) (io.ReadCloser, error)
-	// AddMountUnitFile adds/enables/starts a mount unit.
-	AddMountUnitFile(name, revision, what, where, fstype string) (string, error)
-	// AddMountUnitFileWithOptions adds/enables/starts a mount unit with options.
-	AddMountUnitFileWithOptions(unitOptions *MountUnitOptions) (string, error)
+	// EnsureMountUnitFile adds/enables/starts a mount unit.
+	EnsureMountUnitFile(description, what, where, fstype string, flags EnsureMountUnitFlags) (string, error)
+	// EnsureMountUnitFileWithOptions adds/enables/starts a mount unit with options.
+	EnsureMountUnitFileWithOptions(unitOptions *MountUnitOptions) (string, error)
 	// RemoveMountUnitFile unmounts/stops/disables/removes a mount unit.
 	RemoveMountUnitFile(baseDir string) error
 	// ListMountUnits gets the list of targets of the mount units created by
 	// the `origin` module for the given snap
 	ListMountUnits(snapName, origin string) ([]string, error)
+	// ListUnits get the list of units currently in memory including transient units.
+	ListUnits(pattern string) ([]string, error)
 	// Mask the given service.
 	Mask(service string) error
 	// Unmask the given service.
@@ -405,6 +449,8 @@ type Systemd interface {
 	CurrentTasksCount(unit string) (uint64, error)
 	// Run a command
 	Run(command []string, opts *RunOptions) ([]byte, error)
+	// Set log level for the system
+	SetLogLevel(logLevel string) error
 }
 
 // KeyringMode describes how the kernel keyring is setup, see systemd.exec(5)
@@ -422,22 +468,23 @@ type RunOptions struct {
 	//      and let the caller do the keyring setup but feels a bit loose
 	KeyringMode KeyringMode
 	Stdin       io.Reader
+	Properties  []string
 }
 
 // A Log is a single entry in the systemd journal.
 // In almost all cases, the strings map to a single string value, but as per the
 // manpage for journalctl, under the json format,
 //
-//    Journal entries permit non-unique fields within the same log entry. JSON
-//    does not allow non-unique fields within objects. Due to this, if a
-//    non-unique field is encountered a JSON array is used as field value,
-//    listing all field values as elements.
+//	Journal entries permit non-unique fields within the same log entry. JSON
+//	does not allow non-unique fields within objects. Due to this, if a
+//	non-unique field is encountered a JSON array is used as field value,
+//	listing all field values as elements.
 //
 // and this snippet as well,
 //
-//    Fields containing non-printable or non-UTF8 bytes are
-//    encoded as arrays containing the raw bytes individually
-//    formatted as unsigned numbers.
+//	Fields containing non-printable or non-UTF8 bytes are
+//	encoded as arrays containing the raw bytes individually
+//	formatted as unsigned numbers.
 //
 // as such, we sometimes get array values which need to be handled differently,
 // so we manually try to decode the json for each message into different types.
@@ -566,7 +613,7 @@ func (s *systemd) DaemonReexec() error {
 }
 
 func (s *systemd) EnableNoReload(serviceNames []string) error {
-	if 0 == len(serviceNames) {
+	if len(serviceNames) == 0 {
 		return nil
 	}
 	var args []string
@@ -593,7 +640,7 @@ func (s *systemd) Unmask(serviceName string) error {
 }
 
 func (s *systemd) DisableNoReload(serviceNames []string) error {
-	if 0 == len(serviceNames) {
+	if len(serviceNames) == 0 {
 		return nil
 	}
 	var args []string
@@ -1136,11 +1183,11 @@ func (s *systemd) Restart(serviceNames []string) error {
 	return s.Start(serviceNames)
 }
 
-func (s *systemd) RestartAll(serviceName string) error {
+func (s *systemd) RestartNoWaitForStop(services []string) error {
 	if s.mode == GlobalUserMode {
 		panic("cannot call restart with GlobalUserMode")
 	}
-	_, err := s.systemctl("restart", serviceName, "--all")
+	_, err := s.systemctl(append([]string{"restart"}, services...)...)
 	return err
 }
 
@@ -1360,16 +1407,16 @@ func ExistingMountUnitPath(mountPointDir string) string {
 
 var squashfsFsType = squashfs.FsType
 
-// XXX: After=zfs-mount.service is a workaround for LP: #1922293 (a problem
-// with order of mounting most likely related to zfs-linux and/or systemd).
-// XXX: Remove multi-user.target once we are sure it's unnecessary (see the
-// comments in LP: #1983528).
-const mountUnitTemplate = `[Unit]
-Description=Mount unit for {{.SnapName}}
-{{- with .Revision}}, revision {{.}}{{end}}
-{{- with .Origin}} via {{.}}{{end}}
-Before=snapd.service
-After=zfs-mount.service
+// Note that WantedBy=multi-user.target and Before=local-fs.target are
+// only used to allow downgrading to an older version of snapd.
+//
+// We want (see isBeforeDrivers) some snaps and components to be mounted before
+// modules are loaded (that is before systemd-{udevd,modules-load}).
+const snapMountUnitTmpl = `[Unit]
+Description={{.Description}}
+After=snapd.mounts-pre.target
+Before=snapd.mounts.target{{if isBeforeDrivers .MountUnitType}}
+Before=systemd-udevd.service systemd-modules-load.service{{end}}
 
 [Mount]
 What={{.What}}
@@ -1379,35 +1426,58 @@ Options={{join .Options ","}}
 LazyUnmount=yes
 
 [Install]
-WantedBy=default.target multi-user.target
+WantedBy=snapd.mounts.target
+WantedBy=multi-user.target
 {{- with .Origin}}
 X-SnapdOrigin={{.}}
 {{- end}}
 `
 
-var templateFuncs = template.FuncMap{"join": strings.Join}
-var parsedMountUnitTemplate = template.Must(template.New("unit").Funcs(templateFuncs).Parse(mountUnitTemplate))
+func isBeforeDriversLoadMountUnit(mType MountUnitType) bool {
+	return mType == BeforeDriversLoadMountUnit
+}
+
+var templateFuncs = template.FuncMap{"join": strings.Join,
+	"isBeforeDrivers": isBeforeDriversLoadMountUnit}
+var parsedMountUnitTmpl = template.Must(template.New("unit").Funcs(templateFuncs).Parse(snapMountUnitTmpl))
 
 const (
 	snappyOriginModule = "X-SnapdOrigin"
 )
 
-func writeMountUnitFile(u *MountUnitOptions) (mountUnitName string, err error) {
+func ensureMountUnitFile(u *MountUnitOptions) (mountUnitName string, modified mountUpdateStatus, err error) {
 	if u == nil {
-		return "", errors.New("writeMountUnitFile() expects valid mount options")
+		return "", mountUnchanged, errors.New("ensureMountUnitFile() expects valid mount options")
 	}
 
 	mu := MountUnitPathWithLifetime(u.Lifetime, u.Where)
-	outf, err := osutil.NewAtomicFile(mu, 0644, 0, osutil.NoChown, osutil.NoChown)
-	if err != nil {
-		return "", fmt.Errorf("cannot open mount unit file: %v", err)
+	var unitContent bytes.Buffer
+	if err := parsedMountUnitTmpl.Execute(&unitContent, &u); err != nil {
+		return "", mountUnchanged, fmt.Errorf("cannot generate mount unit: %v", err)
 	}
-	defer outf.Cancel()
 
-	if err := parsedMountUnitTemplate.Execute(outf, &u); err != nil {
-		return "", fmt.Errorf("cannot generate mount unit: %v", err)
+	if osutil.FileExists(mu) {
+		modified = mountUpdated
+	} else {
+		modified = mountCreated
 	}
-	return filepath.Base(mu), outf.Commit()
+
+	if err := os.MkdirAll(filepath.Dir(mu), 0755); err != nil {
+		return "", mountUnchanged, fmt.Errorf("cannot create directory %s: %v", filepath.Dir(mu), err)
+	}
+
+	stateErr := osutil.EnsureFileState(mu, &osutil.MemoryFileState{
+		Content: unitContent.Bytes(),
+		Mode:    0644,
+	})
+
+	if stateErr == osutil.ErrSameState {
+		modified = mountUnchanged
+	} else if stateErr != nil {
+		return "", mountUnchanged, stateErr
+	}
+
+	return filepath.Base(mu), modified, nil
 }
 
 func fsMountOptions(fstype string) []string {
@@ -1436,44 +1506,54 @@ func hostFsTypeAndMountOptions(fstype string) (hostFsType string, options []stri
 	return hostFsType, options
 }
 
-func (s *systemd) AddMountUnitFile(snapName, revision, what, where, fstype string) (string, error) {
+func (s *systemd) EnsureMountUnitFile(description, what, where, fstype string, flags EnsureMountUnitFlags) (string, error) {
 	hostFsType, options := hostFsTypeAndMountOptions(fstype)
 	if osutil.IsDirectory(what) {
 		options = append(options, "bind")
 		hostFsType = "none"
 	}
-	return s.AddMountUnitFileWithOptions(&MountUnitOptions{
-		Lifetime: Persistent,
-		SnapName: snapName,
-		Revision: revision,
-		What:     what,
-		Where:    where,
-		Fstype:   hostFsType,
-		Options:  options,
-	})
+	mountOptions := &MountUnitOptions{
+		Lifetime:                 Persistent,
+		Description:              description,
+		What:                     what,
+		Where:                    where,
+		Fstype:                   hostFsType,
+		Options:                  options,
+		PreventRestartIfModified: flags.PreventRestartIfModified,
+	}
+	if flags.StartBeforeDriversLoad {
+		mountOptions.MountUnitType = BeforeDriversLoadMountUnit
+	}
+	return s.EnsureMountUnitFileWithOptions(mountOptions)
 }
 
-func (s *systemd) AddMountUnitFileWithOptions(unitOptions *MountUnitOptions) (string, error) {
+func (s *systemd) EnsureMountUnitFileWithOptions(unitOptions *MountUnitOptions) (string, error) {
 	daemonReloadLock.Lock()
 	defer daemonReloadLock.Unlock()
 
-	mountUnitName, err := writeMountUnitFile(unitOptions)
+	mountUnitName, modified, err := ensureMountUnitFile(unitOptions)
 	if err != nil {
 		return "", err
 	}
+	if modified != mountUnchanged {
+		// we need to do a daemon-reload here to ensure that systemd really
+		// knows about this new mount unit file
+		if err := s.daemonReloadNoLock(); err != nil {
+			return "", err
+		}
 
-	// we need to do a daemon-reload here to ensure that systemd really
-	// knows about this new mount unit file
-	if err := s.daemonReloadNoLock(); err != nil {
-		return "", err
-	}
+		units := []string{mountUnitName}
+		if err := s.EnableNoReload(units); err != nil {
+			return "", err
+		}
 
-	units := []string{mountUnitName}
-	if err := s.EnableNoReload(units); err != nil {
-		return "", err
-	}
-	if err := s.Start(units); err != nil {
-		return "", err
+		// If just modified, some times it is not convenient to restart
+		if modified != mountUpdated || !unitOptions.PreventRestartIfModified {
+			// Start/restart the created or modified unit now
+			if err := s.RestartNoWaitForStop(units); err != nil {
+				return "", err
+			}
+		}
 	}
 
 	return mountUnitName, nil
@@ -1616,11 +1696,33 @@ func (s *systemd) ListMountUnits(snapName, origin string) ([]string, error) {
 	return mountPoints, nil
 }
 
-func (s *systemd) ReloadOrRestart(serviceName string) error {
+func (s *systemd) ListUnits(pattern string) ([]string, error) {
+	out, err := s.systemctl("list-units", "--output=json", pattern)
+	if err != nil {
+		return nil, err
+	}
+
+	type rawUnit struct {
+		UnitName string `json:"unit"`
+	}
+	var parsedUnits []rawUnit
+	err = json.Unmarshal(out, &parsedUnits)
+	if err != nil {
+		return nil, fmt.Errorf("cannot parse systemctl output: %w", err)
+	}
+
+	units := make([]string, len(parsedUnits))
+	for i, parsedUnit := range parsedUnits {
+		units[i] = parsedUnit.UnitName
+	}
+	return units, nil
+}
+
+func (s *systemd) ReloadOrRestart(serviceNames []string) error {
 	if s.mode == GlobalUserMode {
 		panic("cannot call restart with GlobalUserMode")
 	}
-	_, err := s.systemctl("reload-or-restart", serviceName)
+	_, err := s.systemctl(append([]string{"reload-or-restart"}, serviceNames...)...)
 	return err
 }
 
@@ -1656,17 +1758,40 @@ func (s *systemd) Run(command []string, opts *RunOptions) ([]byte, error) {
 		"--service-type=exec",
 		"--quiet",
 	}
+	if s.mode == UserMode {
+		runArgs = append(runArgs, "--user")
+	}
 	if opts.KeyringMode != "" {
 		runArgs = append(runArgs, fmt.Sprintf("--property=KeyringMode=%v", opts.KeyringMode))
 	}
+	for _, p := range opts.Properties {
+		runArgs = append(runArgs, fmt.Sprintf("--property=%v", p))
+	}
 	runArgs = append(runArgs, "--")
 	runArgs = append(runArgs, command...)
 	cmd := exec.Command("systemd-run", runArgs...)
 	cmd.Stdin = opts.Stdin
 
-	output, err := cmd.CombinedOutput()
+	stdout, stderr, err := osutil.RunCmd(cmd)
 	if err != nil {
-		return nil, fmt.Errorf("cannot run %q: %v", command, osutil.OutputErr(output, err))
+		return nil, fmt.Errorf("cannot run %q: %v", command, osutil.OutputErrCombine(stdout, stderr, err))
 	}
-	return output, nil
+	return stdout, nil
+}
+
+func (s *systemd) SetLogLevel(logLevel string) error {
+	_, err := s.systemctl("log-level", logLevel)
+
+	// Older systemd versions used systemd-analyze instead, try that if error
+	if err != nil {
+		if stdout, stderr, err2 := osutil.RunSplitOutput(
+			"systemd-analyze", "set-log-level", logLevel); err2 == nil {
+			return nil
+		} else {
+			logger.Noticef("while running systemd-analyze: %v",
+				osutil.OutputErrCombine(stdout, stderr, err2))
+		}
+	}
+
+	return err
 }
