@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2020 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,25 +20,28 @@
 package store
 
 import (
+	"context"
 	"encoding/base64"
 	"encoding/json"
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"net"
 	"net/http"
+	"net/url"
+	"os"
 	"path/filepath"
+	"regexp"
+	"strconv"
 	"strings"
 	"time"
 
-	"gopkg.in/tylerb/graceful.v1"
-
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/asserts/sysdb"
 	"github.com/snapcore/snapd/asserts/systestkeys"
-	"github.com/snapcore/snapd/httputil"
-	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/snapfile"
+	"github.com/snapcore/snapd/snapdenv"
 	"github.com/snapcore/snapd/store"
 )
 
@@ -64,7 +67,7 @@ type Store struct {
 	assertFallback bool
 	fallback       *store.Store
 
-	srv *graceful.Server
+	srv *http.Server
 }
 
 // NewStore creates a new store server serving snaps from the given top directory and assertions from topDir/asserts. If assertFallback is true missing assertions are looked up in the main online store.
@@ -72,7 +75,7 @@ func NewStore(topDir, addr string, assertFallback bool) *Store {
 	mux := http.NewServeMux()
 	var sto *store.Store
 	if assertFallback {
-		httputil.SetUserAgentFromVersion("unknown", "fakestore")
+		snapdenv.SetUserAgentFromVersion("unknown", nil, "fakestore")
 		sto = store.New(nil, nil)
 	}
 	store := &Store{
@@ -83,13 +86,9 @@ func NewStore(topDir, addr string, assertFallback bool) *Store {
 		fallback:       sto,
 
 		url: fmt.Sprintf("http://%s", addr),
-		srv: &graceful.Server{
-			Timeout: 2 * time.Second,
-
-			Server: &http.Server{
-				Addr:    addr,
-				Handler: mux,
-			},
+		srv: &http.Server{
+			Addr:    addr,
+			Handler: mux,
 		},
 	}
 
@@ -98,7 +97,11 @@ func NewStore(topDir, addr string, assertFallback bool) *Store {
 	mux.HandleFunc("/api/v1/snaps/details/", store.detailsEndpoint)
 	mux.HandleFunc("/api/v1/snaps/metadata", store.bulkEndpoint)
 	mux.Handle("/download/", http.StripPrefix("/download/", http.FileServer(http.Dir(topDir))))
-	mux.HandleFunc("/api/v1/snaps/assertions/", store.assertionsEndpoint)
+	// v2
+	mux.HandleFunc("/v2/assertions/", store.assertionsEndpoint)
+	mux.HandleFunc("/v2/snaps/refresh", store.snapActionEndpoint)
+
+	mux.HandleFunc("/v2/repairs/", store.repairsEndpoint)
 
 	return store
 }
@@ -119,6 +122,8 @@ func (s *Store) Start() error {
 		return err
 	}
 
+	s.url = fmt.Sprintf("http://%s", l.Addr())
+
 	go s.srv.Serve(l)
 	return nil
 }
@@ -126,12 +131,12 @@ func (s *Store) Start() error {
 // Stop stops the server
 func (s *Store) Stop() error {
 	timeoutTime := 2000 * time.Millisecond
-	s.srv.Stop(timeoutTime / 2)
-
-	select {
-	case <-s.srv.StopChan():
-	case <-time.After(timeoutTime):
-		return fmt.Errorf("store failed to stop after %s", timeoutTime)
+	ctx, cancel := context.WithTimeout(context.Background(), timeoutTime)
+	defer cancel()
+	if err := s.srv.Shutdown(ctx); err != nil {
+		// forceful close
+		s.srv.Close()
+		return fmt.Errorf("store failed to stop after: %s", timeoutTime)
 	}
 
 	return nil
@@ -169,32 +174,38 @@ type essentialInfo struct {
 	Version     string
 	Size        uint64
 	Digest      string
+	Confinement string
+	Type        string
+	Base        string
 }
 
 var errInfo = errors.New("cannot get info")
 
 func snapEssentialInfo(w http.ResponseWriter, fn, snapID string, bs asserts.Backstore) (*essentialInfo, error) {
-	snapFile, err := snap.Open(fn)
+	f, err := snapfile.Open(fn)
 	if err != nil {
 		http.Error(w, fmt.Sprintf("cannot read: %v: %v", fn, err), 400)
 		return nil, errInfo
 	}
 
-	info, err := snap.ReadInfoFromSnapFile(snapFile, nil)
+	restoreSanitize := snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {})
+	defer restoreSanitize()
+
+	info, err := snap.ReadInfoFromSnapFile(f, nil)
 	if err != nil {
-		http.Error(w, fmt.Sprintf("can get info for: %v: %v", fn, err), 400)
+		http.Error(w, fmt.Sprintf("cannot get info for: %v: %v", fn, err), 400)
 		return nil, errInfo
 	}
 
 	snapDigest, size, err := asserts.SnapFileSHA3_384(fn)
 	if err != nil {
-		http.Error(w, fmt.Sprintf("can get digest for: %v: %v", fn, err), 400)
+		http.Error(w, fmt.Sprintf("cannot get digest for: %v: %v", fn, err), 400)
 		return nil, errInfo
 	}
 
 	snapRev, devAcct, err := findSnapRevision(snapDigest, bs)
-	if err != nil && err != asserts.ErrNotFound {
-		http.Error(w, fmt.Sprintf("can get info for: %v: %v", fn, err), 400)
+	if err != nil && !errors.Is(err, &asserts.NotFoundError{}) {
+		http.Error(w, fmt.Sprintf("cannot get info for: %v: %v", fn, err), 400)
 		return nil, errInfo
 	}
 
@@ -213,7 +224,7 @@ func snapEssentialInfo(w http.ResponseWriter, fn, snapID string, bs asserts.Back
 	}
 
 	return &essentialInfo{
-		Name:        info.Name(),
+		Name:        info.SnapName(),
 		SnapID:      snapID,
 		DeveloperID: develID,
 		DevelName:   devel,
@@ -221,17 +232,12 @@ func snapEssentialInfo(w http.ResponseWriter, fn, snapID string, bs asserts.Back
 		Version:     info.Version,
 		Digest:      snapDigest,
 		Size:        size,
+		Confinement: string(info.Confinement),
+		Type:        string(info.Type()),
+		Base:        info.Base,
 	}, nil
 }
 
-type searchPayloadJSON struct {
-	Packages []detailsReplyJSON `json:"clickindex:package"`
-}
-
-type searchReplyJSON struct {
-	Payload searchPayloadJSON `json:"_embedded"`
-}
-
 type detailsReplyJSON struct {
 	Architectures   []string `json:"architecture"`
 	SnapID          string   `json:"snap_id"`
@@ -243,6 +249,9 @@ type detailsReplyJSON struct {
 	Version         string   `json:"version"`
 	Revision        int      `json:"revision"`
 	DownloadDigest  string   `json:"download_sha3_384"`
+	Confinement     string   `json:"confinement"`
+	Type            string   `json:"type"`
+	Base            string   `json:"base,omitempty"`
 }
 
 func (s *Store) searchEndpoint(w http.ResponseWriter, req *http.Request) {
@@ -250,6 +259,84 @@ func (s *Store) searchEndpoint(w http.ResponseWriter, req *http.Request) {
 	fmt.Fprintf(w, "search not implemented")
 }
 
+func (s *Store) repairsEndpoint(w http.ResponseWriter, req *http.Request) {
+	brandAndRepairID := strings.Split(strings.TrimPrefix(req.URL.Path, "/v2/repairs/"), "/")
+	if len(brandAndRepairID) != 2 {
+		http.Error(w, "missing brand and repair ID", 400)
+		return
+	}
+
+	bs, err := s.collectAssertions()
+	if err != nil {
+		http.Error(w, fmt.Sprintf("internal error collecting assertions: %v", err), 500)
+		return
+	}
+
+	a, err := s.retrieveAssertion(bs, asserts.RepairType, brandAndRepairID)
+	if errors.Is(err, &asserts.NotFoundError{}) {
+		w.Header().Set("Content-Type", "application/problem+json")
+		w.WriteHeader(404)
+		w.Write([]byte(`{"status": 404}`))
+		return
+	}
+	if err != nil {
+		http.Error(w, fmt.Sprintf("cannot retrieve repair assertion %v: %v", brandAndRepairID, err), 400)
+		return
+	}
+
+	// handle If-None-Match caching
+	revNumString := req.Header.Get("If-None-Match")
+	if revNumString != "" {
+		revRegexp := regexp.MustCompile(`^"([0-9]+)"$`)
+		match := revRegexp.FindStringSubmatch(revNumString)
+		if match == nil || len(match) != 2 {
+			http.Error(w, fmt.Sprintf("malformed If-None-Match header (%q): must be repair revision number in quotes", revNumString), 400)
+			return
+		}
+		revNum, err := strconv.Atoi(match[1])
+		if err != nil {
+			http.Error(w, fmt.Sprintf("malformed If-None-Match header (%q): %v", revNumString, err), 400)
+			return
+		}
+
+		if revNum == a.Revision() {
+			// if the If-None-Match header is the assertion revision verbatim
+			// then return 304 (Not Modified) and stop
+			w.WriteHeader(304)
+			return
+		}
+	}
+
+	// there are two cases, one where we are asked for the full assertion, and
+	// one where we are asked for JSON headers of the assertion, so check which
+	// one we were asked for by inspecting the Accept header
+	switch accept := req.Header.Get("Accept"); accept {
+	case "application/json":
+		// headers only
+		headers := a.Headers()
+		// we have to wrap the headers in a JSON object under the key
+		// "headers"
+		resp := map[string]interface{}{
+			"headers": headers,
+		}
+		b, err := json.Marshal(resp)
+		if err != nil {
+			http.Error(w, fmt.Sprintf("internal error collecting assertion headers as json: %v", err), 500)
+			return
+		}
+		w.Header().Set("Content-Type", "application/json")
+		w.WriteHeader(200)
+		w.Write(b)
+	case "application/x.ubuntu.assertion":
+		// full assertion
+		w.Header().Set("Content-Type", asserts.MediaType)
+		w.WriteHeader(200)
+		w.Write(asserts.Encode(a))
+	default:
+		http.Error(w, fmt.Sprintf("unsupported Accept format (%q): only application/json and application/x.ubuntu.assertion is support", accept), 400)
+	}
+}
+
 func (s *Store) detailsEndpoint(w http.ResponseWriter, req *http.Request) {
 	pkg := strings.TrimPrefix(req.URL.Path, "/api/v1/snaps/details/")
 	if pkg == req.URL.Path {
@@ -292,6 +379,9 @@ func (s *Store) detailsEndpoint(w http.ResponseWriter, req *http.Request) {
 		Version:         essInfo.Version,
 		Revision:        essInfo.Revision,
 		DownloadDigest:  hexify(essInfo.Digest),
+		Confinement:     essInfo.Confinement,
+		Type:            essInfo.Type,
+		Base:            essInfo.Base,
 	}
 
 	// use indent because this is a development tool, output
@@ -312,16 +402,20 @@ func (s *Store) collectSnaps() (map[string]string, error) {
 
 	snaps := map[string]string{}
 
+	restoreSanitize := snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {})
+	defer restoreSanitize()
+
 	for _, fn := range snapFns {
-		snapFile, err := snap.Open(fn)
+		f, err := snapfile.Open(fn)
 		if err != nil {
 			return nil, err
 		}
-		info, err := snap.ReadInfoFromSnapFile(snapFile, nil)
+		info, err := snap.ReadInfoFromSnapFile(f, nil)
 		if err != nil {
 			return nil, err
 		}
-		snaps[info.Name()] = fn
+		snaps[info.SnapName()] = fn
+		logger.Debugf("found snap %q at %v", info.SnapName(), fn)
 	}
 
 	return snaps, err
@@ -379,7 +473,7 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 	}
 
 	var remoteStore string
-	if osutil.GetenvBool("SNAPPY_USE_STAGING_STORE") {
+	if snapdenv.UseStagingStore() {
 		remoteStore = "staging"
 	} else {
 		remoteStore = "production"
@@ -398,10 +492,9 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 
 	// check if we have downloadable snap of the given SnapID
 	for _, pkg := range pkgs.CandidateSnaps {
-
 		name := snapIDtoName[pkg.SnapID]
 		if name == "" {
-			http.Error(w, fmt.Sprintf("unknown snapid: %q", pkg.SnapID), 400)
+			http.Error(w, fmt.Sprintf("unknown snap-id: %q", pkg.SnapID), 400)
 			return
 		}
 
@@ -425,6 +518,9 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 				Version:         essInfo.Version,
 				Revision:        essInfo.Revision,
 				DownloadDigest:  hexify(essInfo.Digest),
+				Confinement:     essInfo.Confinement,
+				Type:            essInfo.Type,
+				Base:            essInfo.Base,
 			})
 		}
 	}
@@ -433,7 +529,7 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 	// should look nice
 	out, err := json.MarshalIndent(replyData, "", "    ")
 	if err != nil {
-		http.Error(w, fmt.Sprintf("can marshal: %v: %v", replyData, err), 400)
+		http.Error(w, fmt.Sprintf("cannot marshal: %v: %v", replyData, err), 400)
 		return
 	}
 	w.Write(out)
@@ -444,7 +540,9 @@ func (s *Store) collectAssertions() (asserts.Backstore, error) {
 	bs := asserts.NewMemoryBackstore()
 
 	add := func(a asserts.Assertion) {
-		bs.Put(a.Type(), a)
+		if err := bs.Put(a.Type(), a); err != nil {
+			logger.Noticef("cannot add assertion %q: %v", a.Headers(), err)
+		}
 	}
 
 	for _, t := range sysdb.Trusted() {
@@ -460,7 +558,7 @@ func (s *Store) collectAssertions() (asserts.Backstore, error) {
 	}
 
 	for _, fn := range aFiles {
-		b, err := ioutil.ReadFile(fn)
+		b, err := os.ReadFile(fn)
 		if err != nil {
 			return nil, err
 		}
@@ -476,26 +574,66 @@ func (s *Store) collectAssertions() (asserts.Backstore, error) {
 	return bs, nil
 }
 
-func isAssertNotFound(err error) bool {
-	if err == asserts.ErrNotFound {
-		return true
-	}
-	if _, ok := err.(*store.AssertionNotFoundError); ok {
-		return true
-	}
-	return false
+type currentSnap struct {
+	SnapID      string `json:"snap-id"`
+	InstanceKey string `json:"instance-key"`
 }
 
-func (s *Store) retrieveAssertion(bs asserts.Backstore, assertType *asserts.AssertionType, primaryKey []string) (asserts.Assertion, error) {
-	a, err := bs.Get(assertType, primaryKey, assertType.MaxSupportedFormat())
-	if err == asserts.ErrNotFound && s.assertFallback {
-		return s.fallback.Assertion(assertType, primaryKey, nil)
-	}
-	return a, err
+type snapAction struct {
+	Action      string `json:"action"`
+	InstanceKey string `json:"instance-key"`
+	SnapID      string `json:"snap-id"`
+	Name        string `json:"name"`
+	Revision    int    `json:"revision,omitempty"`
 }
 
-func (s *Store) assertionsEndpoint(w http.ResponseWriter, req *http.Request) {
-	assertPath := strings.TrimPrefix(req.URL.Path, "/api/v1/snaps/assertions/")
+type snapActionRequest struct {
+	Context []currentSnap `json:"context"`
+	Fields  []string      `json:"fields"`
+	Actions []snapAction  `json:"actions"`
+}
+
+type snapActionResult struct {
+	Result      string          `json:"result"`
+	InstanceKey string          `json:"instance-key"`
+	SnapID      string          `json:"snap-id"`
+	Name        string          `json:"name"`
+	Snap        detailsResultV2 `json:"snap"`
+}
+
+type snapActionResultList struct {
+	Results []*snapActionResult `json:"results"`
+}
+
+type detailsResultV2 struct {
+	Architectures []string `json:"architectures"`
+	Base          string   `json:"base,omitempty"`
+	SnapID        string   `json:"snap-id"`
+	Name          string   `json:"name"`
+	Publisher     struct {
+		ID       string `json:"id"`
+		Username string `json:"username"`
+	} `json:"publisher"`
+	Download struct {
+		URL      string `json:"url"`
+		Sha3_384 string `json:"sha3-384"`
+		Size     uint64 `json:"size"`
+	} `json:"download"`
+	Version     string `json:"version"`
+	Revision    int    `json:"revision"`
+	Confinement string `json:"confinement"`
+	Type        string `json:"type"`
+}
+
+func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
+	var reqData snapActionRequest
+	var replyData snapActionResultList
+
+	decoder := json.NewDecoder(req.Body)
+	if err := decoder.Decode(&reqData); err != nil {
+		http.Error(w, fmt.Sprintf("cannot decode request body: %v", err), 400)
+		return
+	}
 
 	bs, err := s.collectAssertions()
 	if err != nil {
@@ -503,40 +641,198 @@ func (s *Store) assertionsEndpoint(w http.ResponseWriter, req *http.Request) {
 		return
 	}
 
-	comps := strings.Split(assertPath, "/")
-
-	if len(comps) == 0 {
-		http.Error(w, "missing assertion type", 400)
+	var remoteStore string
+	if snapdenv.UseStagingStore() {
+		remoteStore = "staging"
+	} else {
+		remoteStore = "production"
+	}
+	snapIDtoName, err := addSnapIDs(bs, someSnapIDtoName[remoteStore])
+	if err != nil {
+		http.Error(w, fmt.Sprintf("internal error collecting snapIDs: %v", err), 500)
 		return
+	}
 
+	snaps, err := s.collectSnaps()
+	if err != nil {
+		http.Error(w, fmt.Sprintf("internal error collecting snaps: %v", err), 500)
+		return
+	}
+
+	actions := reqData.Actions
+	if len(actions) == 1 && actions[0].Action == "refresh-all" {
+		actions = make([]snapAction, len(reqData.Context))
+		for i, s := range reqData.Context {
+			actions[i] = snapAction{
+				Action:      "refresh",
+				SnapID:      s.SnapID,
+				InstanceKey: s.InstanceKey,
+			}
+		}
+	}
+
+	// check if we have downloadable snap of the given SnapID or name
+	for _, a := range actions {
+		name := a.Name
+		snapID := a.SnapID
+		if a.Action == "refresh" {
+			name = snapIDtoName[snapID]
+		}
+
+		if name == "" {
+			http.Error(w, fmt.Sprintf("unknown snap-id: %q", snapID), 400)
+			return
+		}
+
+		if fn, ok := snaps[name]; ok {
+			essInfo, err := snapEssentialInfo(w, fn, snapID, bs)
+			if essInfo == nil {
+				if err != errInfo {
+					panic(err)
+				}
+				return
+			}
+
+			res := &snapActionResult{
+				Result:      a.Action,
+				InstanceKey: a.InstanceKey,
+				SnapID:      essInfo.SnapID,
+				Name:        essInfo.Name,
+				Snap: detailsResultV2{
+					Architectures: []string{"all"},
+					SnapID:        essInfo.SnapID,
+					Name:          essInfo.Name,
+					Version:       essInfo.Version,
+					Revision:      essInfo.Revision,
+					Confinement:   essInfo.Confinement,
+					Type:          essInfo.Type,
+					Base:          essInfo.Base,
+				},
+			}
+			logger.Debugf("requested snap %q revision %d", essInfo.Name, a.Revision)
+			res.Snap.Publisher.ID = essInfo.DeveloperID
+			res.Snap.Publisher.Username = essInfo.DevelName
+			res.Snap.Download.URL = fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn))
+			res.Snap.Download.Sha3_384 = hexify(essInfo.Digest)
+			res.Snap.Download.Size = essInfo.Size
+			replyData.Results = append(replyData.Results, res)
+		}
+	}
+
+	// use indent because this is a development tool, output
+	// should look nice
+	out, err := json.MarshalIndent(replyData, "", "    ")
+	if err != nil {
+		http.Error(w, fmt.Sprintf("cannot marshal: %v: %v", replyData, err), 400)
+		return
+	}
+	w.Write(out)
+}
+
+func (s *Store) retrieveAssertion(bs asserts.Backstore, assertType *asserts.AssertionType, primaryKey []string) (asserts.Assertion, error) {
+	a, err := bs.Get(assertType, primaryKey, assertType.MaxSupportedFormat())
+	if errors.Is(err, &asserts.NotFoundError{}) && s.assertFallback {
+		return s.fallback.Assertion(assertType, primaryKey, nil)
+	}
+	return a, err
+}
+
+func (s *Store) retrieveLatestSequenceFormingAssertion(bs asserts.Backstore, assertType *asserts.AssertionType, sequenceKey []string) (asserts.Assertion, error) {
+	a, err := bs.SequenceMemberAfter(assertType, sequenceKey, -1, assertType.MaxSupportedFormat())
+	if errors.Is(err, &asserts.NotFoundError{}) && s.assertFallback {
+		return s.fallback.SeqFormingAssertion(assertType, sequenceKey, -1, nil)
+	}
+	return a, err
+}
+
+func (s *Store) sequenceFromQueryValues(values url.Values) (int, error) {
+	if val, ok := values["sequence"]; ok {
+		// special case value of 'latest', in that case
+		// we return -1 to indicate we want the newest
+		if val[0] != "latest" {
+			seq, err := strconv.Atoi(val[0])
+			if err != nil {
+				return -1, fmt.Errorf("cannot parse sequence %s: %v", val[0], err)
+			}
+
+			// Only positive integers and 'latest' are valid
+			if seq <= 0 {
+				return -1, fmt.Errorf("the requested sequence must be above 0")
+			}
+			return seq, nil
+		}
+	}
+	return -1, nil
+}
+
+func (s *Store) assertTypeAndKey(urlPath string) (*asserts.AssertionType, []string, error) {
+	// trim the assertions prefix, and handle any query parameters
+	assertPath := strings.TrimPrefix(urlPath, "/v2/assertions/")
+	comps := strings.Split(assertPath, "/")
+	if len(comps) == 0 {
+		return nil, nil, fmt.Errorf("missing assertion type")
 	}
 
 	typ := asserts.Type(comps[0])
 	if typ == nil {
-		http.Error(w, fmt.Sprintf("unknown assertion type: %s", comps[0]), 400)
+		return nil, nil, fmt.Errorf("unknown assertion type: %s", comps[0])
+	}
+	return typ, comps[1:], nil
+}
+
+func (s *Store) retrieveAssertionWrapper(bs asserts.Backstore, assertType *asserts.AssertionType, keyParts []string, values url.Values) (asserts.Assertion, error) {
+	pk := keyParts
+	if assertType.SequenceForming() {
+		seq, err := s.sequenceFromQueryValues(values)
+		if err != nil {
+			return nil, err
+		}
+
+		// If no sequence value was provided, or when requesting the latest sequence
+		// point of an assertion, we use a different method of resolving the assertion.
+		if seq <= 0 {
+			return s.retrieveLatestSequenceFormingAssertion(bs, assertType, keyParts)
+		}
+
+		// Otherwise append the sequence to form the primary key and use
+		// the default retrieval.
+		pk = append(pk, strconv.Itoa(seq))
+	}
+
+	if !assertType.AcceptablePrimaryKey(pk) {
+		return nil, fmt.Errorf("wrong primary key length: %v", pk)
+	}
+	return s.retrieveAssertion(bs, assertType, pk)
+}
+
+func (s *Store) assertionsEndpoint(w http.ResponseWriter, req *http.Request) {
+	typ, pk, err := s.assertTypeAndKey(req.URL.Path)
+	if err != nil {
+		http.Error(w, err.Error(), 400)
 		return
 	}
 
-	if len(typ.PrimaryKey) != len(comps)-1 {
-		http.Error(w, fmt.Sprintf("wrong primary key length: %v", comps), 400)
+	bs, err := s.collectAssertions()
+	if err != nil {
+		http.Error(w, fmt.Sprintf("internal error collecting assertions: %v", err), 500)
 		return
 	}
 
-	a, err := s.retrieveAssertion(bs, typ, comps[1:])
-	if isAssertNotFound(err) {
+	as, err := s.retrieveAssertionWrapper(bs, typ, pk, req.URL.Query())
+	if errors.Is(err, &asserts.NotFoundError{}) {
 		w.Header().Set("Content-Type", "application/problem+json")
 		w.WriteHeader(404)
-		w.Write([]byte(`{"status": 404}`))
+		w.Write([]byte(`{"error-list":[{"code":"not-found","message":"not found"}]}`))
 		return
 	}
 	if err != nil {
-		http.Error(w, fmt.Sprintf("cannot retrieve assertion %v: %v", comps, err), 400)
+		http.Error(w, fmt.Sprintf("cannot retrieve assertion %v: %v", pk, err), 400)
 		return
 	}
 
 	w.Header().Set("Content-Type", asserts.MediaType)
 	w.WriteHeader(200)
-	w.Write(asserts.Encode(a))
+	w.Write(asserts.Encode(as))
 }
 
 func addSnapIDs(bs asserts.Backstore, initial map[string]string) (map[string]string, error) {
