@@ -22,10 +22,13 @@ package store
 import (
 	"bytes"
 	"encoding/json"
-	"io/ioutil"
+	"errors"
+	"io"
 	"net/http"
+	"net/url"
 	"os"
 	"path/filepath"
+	"syscall"
 	"testing"
 	"text/template"
 
@@ -47,8 +50,6 @@ type storeTestSuite struct {
 
 var _ = Suite(&storeTestSuite{})
 
-var defaultAddr = "localhost:23321"
-
 func getSha(fn string) (string, uint64) {
 	snapDigest, size, err := asserts.SnapFileSHA3_384(fn)
 	if err != nil {
@@ -61,7 +62,7 @@ func (s *storeTestSuite) SetUpTest(c *C) {
 	topdir := c.MkDir()
 	err := os.Mkdir(filepath.Join(topdir, "asserts"), 0755)
 	c.Assert(err, IsNil)
-	s.store = NewStore(topdir, defaultAddr, false)
+	s.store = NewStore(topdir, "localhost:0", false)
 	err = s.store.Start()
 	c.Assert(err, IsNil)
 
@@ -88,7 +89,22 @@ func (s *storeTestSuite) StorePostJSON(path string, content []byte) (*http.Respo
 }
 
 func (s *storeTestSuite) TestStoreURL(c *C) {
-	c.Assert(s.store.URL(), Equals, "http://"+defaultAddr)
+	u, err := url.Parse(s.store.URL())
+	c.Assert(err, IsNil)
+	c.Check(u.Hostname(), Equals, "127.0.0.1")
+	c.Check(u.Port(), Not(Equals), "")
+}
+
+func (s *storeTestSuite) TestStoreListenAddr(c *C) {
+	u, err := url.Parse(s.store.URL())
+	c.Assert(err, IsNil)
+	// use the same address as the fake store started by the tests so that
+	// this will fail with addr-in-use error
+	topdir := c.MkDir()
+	newstore := NewStore(topdir, u.Host, false)
+	err = newstore.Start()
+	c.Assert(err, NotNil)
+	c.Check(errors.Is(err, syscall.EADDRINUSE), Equals, true)
 }
 
 func (s *storeTestSuite) TestTrivialGetWorks(c *C) {
@@ -97,7 +113,7 @@ func (s *storeTestSuite) TestTrivialGetWorks(c *C) {
 	defer resp.Body.Close()
 
 	c.Assert(resp.StatusCode, Equals, 418)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Assert(string(body), Equals, "I'm a teapot")
 
@@ -109,7 +125,7 @@ func (s *storeTestSuite) TestSearchEndpoint(c *C) {
 	defer resp.Body.Close()
 
 	c.Assert(resp.StatusCode, Equals, 501)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Assert(string(body), Equals, "search not implemented")
 
@@ -471,7 +487,7 @@ func (s *storeTestSuite) TestAssertionsEndpointPreloaded(c *C) {
 	c.Assert(resp.StatusCode, Equals, 200)
 	c.Check(resp.Header.Get("Content-Type"), Equals, "application/x.ubuntu.assertion")
 
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, string(asserts.Encode(systestkeys.TestRootAccount)))
 }
@@ -482,7 +498,7 @@ func (s *storeTestSuite) TestAssertionsEndpointFromAssertsDir(c *C) {
 	c.Assert(err, IsNil)
 	rev := a.(*asserts.SnapRevision)
 
-	err = ioutil.WriteFile(filepath.Join(s.store.assertDir, "foo_36.snap-revision"), []byte(exampleSnapRev), 0655)
+	err = os.WriteFile(filepath.Join(s.store.assertDir, "foo_36.snap-revision"), []byte(exampleSnapRev), 0655)
 	c.Assert(err, IsNil)
 
 	resp, err := s.StoreGet(`/v2/assertions/snap-revision/` + rev.SnapSHA3_384())
@@ -490,13 +506,13 @@ func (s *storeTestSuite) TestAssertionsEndpointFromAssertsDir(c *C) {
 	defer resp.Body.Close()
 
 	c.Assert(resp.StatusCode, Equals, 200)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, exampleSnapRev)
 }
 
 func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertion(c *C) {
-	err := ioutil.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
+	err := os.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
 	c.Assert(err, IsNil)
 
 	resp, err := s.StoreGet(`/v2/assertions/validation-set/16/canonical/base-set?sequence=2`)
@@ -504,13 +520,13 @@ func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertion(c *C) {
 	defer resp.Body.Close()
 
 	c.Check(resp.StatusCode, Equals, 200)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, exampleValidationSet)
 }
 
 func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertionLatest(c *C) {
-	err := ioutil.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
+	err := os.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
 	c.Assert(err, IsNil)
 
 	resp, err := s.StoreGet(`/v2/assertions/validation-set/16/canonical/base-set?sequence=latest`)
@@ -518,13 +534,13 @@ func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertionLatest(c *C) {
 	defer resp.Body.Close()
 
 	c.Check(resp.StatusCode, Equals, 200)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, exampleValidationSet)
 }
 
 func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertionInvalidSequence(c *C) {
-	err := ioutil.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
+	err := os.WriteFile(filepath.Join(s.store.assertDir, "base-set.validation-set"), []byte(exampleValidationSet), 0655)
 	c.Assert(err, IsNil)
 
 	resp, err := s.StoreGet(`/v2/assertions/validation-set/16/canonical/base-set?sequence=0`)
@@ -532,7 +548,7 @@ func (s *storeTestSuite) TestAssertionsEndpointSequenceAssertionInvalidSequence(
 	defer resp.Body.Close()
 
 	c.Assert(resp.StatusCode, Equals, 400)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, "cannot retrieve assertion [16 canonical base-set]: the requested sequence must be above 0\n")
 }
@@ -543,7 +559,7 @@ func (s *storeTestSuite) TestAssertionsEndpointSequenceInvalid(c *C) {
 	defer resp.Body.Close()
 
 	c.Assert(resp.StatusCode, Equals, 400)
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	c.Assert(err, IsNil)
 	c.Check(string(body), Equals, "cannot retrieve assertion [16 canonical base-set]: cannot parse sequence foo: strconv.Atoi: parsing \"foo\": invalid syntax\n")
 }
